import abc
from enum import Enum
from typing import Dict, List, Optional, Type, Union

from pydantic import PositiveInt, confloat

from geolib.models import BaseDataClass

from ...geometry.one import Point
from ...utils import snake_to_camel
from .internal import (
    AnalysisType,
    NullablePersistablePoint,
    OptionsType,
    PersistableBishopBruteForceSettings,
    PersistableBishopSettings,
    PersistableCircle,
    PersistableGeneticSlipPlaneConstraints,
    PersistableGridEnhancements,
    PersistablePoint,
    PersistableSearchArea,
    PersistableSearchGrid,
    PersistableSlipPlaneConstraints,
    PersistableSpencerGeneticSettings,
    PersistableSpencerSettings,
    PersistableTangentArea,
    PersistableTangentLines,
    PersistableTwoCirclesOnTangentLine,
    PersistableUpliftVanParticleSwarmSettings,
    PersistableUpliftVanSettings,
)


class DStabilityObject(BaseDataClass, metaclass=abc.ABCMeta):
    """Base Class for objects in the analysis module."""

    @abc.abstractmethod
    def _to_internal_datastructure(self):
        """Convert to internal datastructure"""
        raise NotImplementedError


class DStabilityAnalysisMethod(DStabilityObject):
    _analysis_type: AnalysisType

    @property
    def analysis_type(self) -> AnalysisType:
        return self._analysis_type


class DStabilityCircle(DStabilityObject):
    """Generates a circle object

    Args:
        center (Point): Center of the circle.
        radius (float): Radius of the circle.
    """

    center: Point
    radius: confloat(gt=0)

    def _to_internal_datastructure(self) -> PersistableCircle:
        return PersistableCircle(
            Center=PersistablePoint(X=self.center.x, Z=self.center.z), Radius=self.radius
        )


class DStabilitySearchGrid(DStabilityObject):
    """Generates a search grid object

    Args:
        bottom_left (Point): Bottom left position of the search grid.
        number_of_points_in_x (int): Number of points to the right
        number_of_points_in_z (int): Number op points upwards
        space (float): Space between the points in x and z direction
    """

    bottom_left: Point
    number_of_points_in_x: PositiveInt
    number_of_points_in_z: PositiveInt
    space: confloat(gt=0)

    def _to_internal_datastructure(self) -> PersistableSearchGrid:
        return PersistableSearchGrid(
            BottomLeft=PersistablePoint(X=self.bottom_left.x, Z=self.bottom_left.z),
            NumberOfPointsInX=self.number_of_points_in_x,
            NumberOfPointsInZ=self.number_of_points_in_z,
            Space=self.space,
        )


class DStabilitySearchArea(DStabilityObject):
    """Generates a search area object

    Args:
        height (float): Height of the search area.
        top_left (Point): Top left position of the search area
        width (float): Width of the search area
    """

    height: confloat(gt=0)
    top_left: Point
    width: confloat(gt=0)

    def _to_internal_datastructure(self) -> PersistableSearchArea:
        return PersistableSearchArea(
            Height=self.height,
            TopLeft=NullablePersistablePoint(X=self.top_left.x, Z=self.top_left.z),
            Width=self.width,
        )


class DStabilitySlipPlaneConstraints(DStabilityObject):
    """Generates a slip plane constraints object

    Args:
        is_size_constraints_enabled (bool): enabel size constraints, defaults to False
        is_zone_a_constraints_enabled (bool): enable constraints for zone A, defaults to False
        is_zone_b_constraints_enabled (bool): enable constraints for zone B, defaults to False
        minimum_slip_plane_depth (float): minimum slip plane depth, defaults to 0.0
        minimum_slip_plane_length (float): minimum slip plane length, defaults to 0.0
        width_zone_a (float): width of zone A, defaults to 0.0
        width_zone_b (float): width of zone B, defaults to 0.0
        x_left_zone_a (float): x coordinate of left point of zone A, defaults to 0.0
        x_left_zone_b (float): x coordinate of left point of zone B, defaults to 0.0
    """

    is_size_constraints_enabled: bool = False
    is_zone_a_constraints_enabled: bool = False
    is_zone_b_constraints_enabled: bool = False
    minimum_slip_plane_depth: float = 0.0
    minimum_slip_plane_length: float = 0.0
    width_zone_a: float = 0.0
    width_zone_b: float = 0.0
    x_left_zone_a: float = 0.0
    x_left_zone_b: float = 0.0

    def _to_internal_datastructure(self) -> PersistableSlipPlaneConstraints:
        data = {**{snake_to_camel(name): value for name, value in self.dict().items()}}
        return PersistableSlipPlaneConstraints(**data)


class DStabilityGeneticSlipPlaneConstraints(DStabilityObject):
    """Generates slip plane constraints object for the Spencer method

    Args:
        is_enabled (bool): Height of the search area.
        minimum_angle_between_slices (float): Minimum angle between slices, defaults to 0.
        minimum_thrust_line_percentage_inside_slices (float): Minimum thrustline percentage inside slices, defaults to 0.
    """

    is_enabled: bool = False
    minimum_angle_between_slices: float = 0.0
    minimum_thrust_line_percentage_inside_slices: float = 0.0

    def _to_internal_datastructure(self) -> PersistableGeneticSlipPlaneConstraints:
        data = {**{snake_to_camel(name): value for name, value in self.dict().items()}}
        return PersistableGeneticSlipPlaneConstraints(**data)


class DStabilityBishopAnalysisMethod(DStabilityAnalysisMethod):
    """Generates slip plane constraints object for the Bishop method

    Args:
        circle (DStabilityCircle): Bishop circle definition
    """

    _analysis_type: AnalysisType = AnalysisType.BISHOP
    circle: DStabilityCircle

    def _to_internal_datastructure(self) -> PersistableBishopSettings:
        return PersistableBishopSettings(Circle=self.circle._to_internal_datastructure())


class DStabilityBishopBruteForceAnalysisMethod(DStabilityAnalysisMethod):
    """Generates slip plane constraints object for the Bishop Brute Force method

    Args:
        extrapolate_search_space (bool): Extrapolate the search space, defaults to True
        search_grid (DStabilitySearchGrid):
        slip_plane_constraints (DStabilitySlipPlaneConstraints):
        bottom_tangent_line_z (float):
        number_of_tangent_lines (int):
        space_tangent_lines (float):
    """

    _analysis_type: AnalysisType = AnalysisType.BISHOP_BRUTE_FORCE
    extrapolate_search_space: bool = True
    search_grid: DStabilitySearchGrid
    slip_plane_constraints: DStabilitySlipPlaneConstraints = (
        DStabilitySlipPlaneConstraints()
    )
    bottom_tangent_line_z: float
    number_of_tangent_lines: PositiveInt
    space_tangent_lines: confloat(gt=0)

    def _to_internal_datastructure(self) -> PersistableBishopBruteForceSettings:
        return PersistableBishopBruteForceSettings(
            GridEnhancements=PersistableGridEnhancements(
                ExtrapolateSearchSpace=self.extrapolate_search_space
            ),
            SearchGrid=self.search_grid._to_internal_datastructure(),
            SlipPlaneConstraints=self.slip_plane_constraints._to_internal_datastructure(),
            TangentLines=PersistableTangentLines(
                BottomTangentLineZ=self.bottom_tangent_line_z,
                NumberOfTangentLines=self.number_of_tangent_lines,
                Space=self.space_tangent_lines,
            ),
        )


class DStabilitySpencerAnalysisMethod(DStabilityAnalysisMethod):
    """Generates slip plane for the Spencer method

    Args:
        slipplane ([Point]): The points of the slipplane
    """

    _analysis_type: AnalysisType = AnalysisType.SPENCER
    slipplane: List[Point]

    def _to_internal_datastructure(self) -> PersistableSpencerSettings:
        return PersistableSpencerSettings(
            SlipPlane=[PersistablePoint(X=point.x, Z=point.z) for point in self.slipplane]
        )


class DStabilitySpencerGeneticAnalysisMethod(DStabilityAnalysisMethod):
    """Generates the input for the genetic spencer algorithm

    Args:
        options_type (OptionsType): DEFAULT or THOROUGH, defaults to DEFAULT
        slip_plane_a (List[Point]): upper slip plane boundary
        slip_plane_b (List[Point]): lower slip line boundary
        slip_plane_constraints (DStabilityGeneticSlipPlaneConstraints): constraints for the slip plane
    """

    _analysis_type: AnalysisType = AnalysisType.SPENCER_GENETIC
    options_type: OptionsType = OptionsType.DEFAULT
    slip_plane_a: List[Point]
    slip_plane_b: List[Point]
    slip_plane_constraints: DStabilityGeneticSlipPlaneConstraints = (
        DStabilityGeneticSlipPlaneConstraints()
    )

    def _to_internal_datastructure(self) -> PersistableSpencerGeneticSettings:
        return PersistableSpencerGeneticSettings(
            OptionsType=self.options_type,
            SlipPlaneA=[
                PersistablePoint(X=point.x, Z=point.z) for point in self.slip_plane_a
            ],
            SlipPlaneB=[
                PersistablePoint(X=point.x, Z=point.z) for point in self.slip_plane_b
            ],
            SlipPlaneConstraints=self.slip_plane_constraints._to_internal_datastructure(),
        )


class DStabilityUpliftVanAnalysisMethod(DStabilityAnalysisMethod):
    """Generates the input for the uplift van analysis

    Args:
        first_circle (DStabilityCircle): The location of the left circle for the slipplane.
        second_circle_center (Point): The center of the circle on the right side of the slipplane
    """

    _analysis_type: AnalysisType = AnalysisType.UPLIFT_VAN
    first_circle: DStabilityCircle
    second_circle_center: Point

    def _to_internal_datastructure(self) -> PersistableUpliftVanSettings:
        return PersistableUpliftVanSettings(
            SlipPlane=PersistableTwoCirclesOnTangentLine(
                FirstCircleCenter=NullablePersistablePoint(
                    X=self.first_circle.center.x, Z=self.first_circle.center.z
                ),
                FirstCircleRadius=self.first_circle.radius,
                SecondCircleCenter=NullablePersistablePoint(
                    X=self.second_circle_center.x, Z=self.second_circle_center.z
                ),
            )
        )


class DStabilityUpliftVanParticleSwarmAnalysisMethod(DStabilityAnalysisMethod):
    """Generates the input for the uplift van analysis

    Args:
        options_type (OptionsType): DEFAULT or THOROUGH, defaults to DEFAULT
        search_area_a (DStabilitySearchArea): The search area for the circle on the left side of the slip plane
        search_area_b (DStabilitySearchArea): The search area for the circle on the right side of the slip plane
        slip_plane_constraints (DStabilitySlipPlaneConstraints): Slip plane constraints
        tangent_area_height (float): height of the tangent lines search area
        tangent_area_top_z (float): top z coordinate of the tangent lines area
    """

    _analysis_type: AnalysisType = AnalysisType.UPLIFT_VAN_PARTICLE_SWARM
    options_type: OptionsType = OptionsType.DEFAULT
    search_area_a: DStabilitySearchArea
    search_area_b: DStabilitySearchArea
    slip_plane_constraints: DStabilitySlipPlaneConstraints = (
        DStabilitySlipPlaneConstraints()
    )
    tangent_area_height: float
    tangent_area_top_z: float

    def _to_internal_datastructure(self) -> PersistableUpliftVanParticleSwarmSettings:
        return PersistableUpliftVanParticleSwarmSettings(
            OptionsType=self.options_type,
            SearchAreaA=self.search_area_a._to_internal_datastructure(),
            SearchAreaB=self.search_area_b._to_internal_datastructure(),
            SlipPlaneConstraints=self.slip_plane_constraints._to_internal_datastructure(),
            TangentArea=PersistableTangentArea(
                Height=self.tangent_area_height, TopZ=self.tangent_area_top_z
            ),
        )



from pathlib import Path
from typing import Any
from typing import BinaryIO
from typing import Dict
from typing import Union

from pydantic import FilePath

from geolib.models import BaseDataClass


class BaseSerializer(BaseDataClass):
    """Basic class for serializers."""

    ds: Dict[str, Any]

    def render(self) -> str:
        return str(self.ds)

    def write(self, filename: Union[FilePath, BinaryIO]):
        """Write serialized model to Filepath or BytesIO buffer"""
        # if filename is pathlike, open file (in text mode) and write str
        if isinstance(filename, Path):
            with open(filename, "w", encoding='cp1252') as io:
                io.write(self.render())

        # if filename is opened file (in binary mode) or BytesIO object, write render as bytes
        else:
            filename.write(self.render().encode('cp1252'))



import logging
from typing import Set

from geolib.models.validators import BaseValidator

logger = logging.getLogger(__name__)


class DGeoFlowValidator(BaseValidator):
    """Validator for DGeoFlow structure.

    Has access to self.ds from parent class.
    Will run all is_valid_ functions to check for validity."""



from . import Soil

# Clays
soft_clay = Soil(id="soft_clay", name="Soft Clay")
medium_clay = Soil(id="medium_clay", name="Medium Clay")
stiff_clay = Soil(id="stiff_clay", name="Stiff Clay")

# Peats
peat = Soil(id="peat", name="Peat")

# Sands
loose_sand = Soil(id="loose_sand", name="Loose Sand")
stiff_sand = Soil(id="stiff_sand", name="Stiff Sand")
sand = Soil(id="sand", name="Sand")

# Others
gravel = Soil(id="gravel", name="Gravel")
loam = Soil(id="loam", name="Loam")
muck = Soil(id="muck", name="Muck")
undetermined = Soil(id="undetermined", name="Undetermined")



"""
This module handles the three types of state types in DStability.
"""

import abc
from typing import List, Tuple

from geolib.models import BaseDataClass

from ...geometry.one import Point
from ...utils import snake_to_camel
from .internal import (
    PersistablePoint,
    PersistableStateLinePoint,
    PersistableStatePoint,
    PersistableStochasticParameter,
    PersistableStress,
    StateType,
)


class DStabilityObject(BaseDataClass, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def _to_internal_datastructure(self):
        raise NotImplementedError


class DStabilityStress(DStabilityObject):
    """DStability Stress

    Args:
        ocr (float): OCR value, defaults to 1.0
        pop (float): POP value, defaults to 0.0
        stochastic_parameter (PersistableStochasticParameter)
        state_type (StateType): type of state
    """

    ocr: float = 1.0
    pop: float = 0.0
    stochastic_parameter: PersistableStochasticParameter = (
        PersistableStochasticParameter()
    )
    state_type: StateType = StateType.POP

    def _to_internal_datastructure(self) -> PersistableStress:
        data = {
            **{
                snake_to_camel(name): value
                for name, value in self.dict(exclude={"stochastic_parameter"}).items()
            }
        }
        return PersistableStress(**data)


class DStabilityStatePoint(DStabilityObject):
    """DStability StatePoint

    Args:
        id (int): id of the statepoint
        layer_id (int): id of the layer to add the statepoint, note that the API does not check if this point is within the layer
        pop (float): POP value, defaults to 0.0
        point (Point): location of the statepoint
        stress (DStabilityStress): DStabilityStress object
        is_probabilistic (bool): is probabilistic, default to false
        label (str): label of the statepoint

    """

    id: int = -1  # this will be filled in by the datamodel, not meant for the user
    layer_id: int
    point: Point
    stress: DStabilityStress
    is_probabilistic: bool = False
    label: str = ""

    def _to_internal_datastructure(self) -> PersistableStatePoint:
        data = {
            **{
                snake_to_camel(name): value
                for name, value in self.dict().items()
                if name not in {"point", "stress"}
            },
            "Point": PersistablePoint(X=self.point.x, Z=self.point.z),
            "Stress": self.stress._to_internal_datastructure(),
        }
        return PersistableStatePoint(**data)


class DStabilityStateLinePoint(DStabilityObject):
    """"""

    id: int = -1
    above: DStabilityStress
    below: DStabilityStress
    is_above_and_below_correlated: bool = False
    is_probabilistic: bool = False
    label: str = ""
    x: float

    def _to_internal_datastructure(self) -> PersistableStateLinePoint:
        data = {
            **{
                snake_to_camel(name): value
                for name, value in self.dict().items()
                if name not in {"above", "below"}
            },
            "Above": self.above._to_internal_datastructure(),
            "Below": self.below._to_internal_datastructure(),
        }
        return PersistableStateLinePoint(**data)



# FROM https://github.com/python/cpython/blob/6292be7adf247589bbf03524f8883cb4cb61f3e9/Lib/typing.py
import collections
import sys
from typing import Dict, List, Tuple, Type, _GenericAlias, _SpecialForm, get_type_hints

if sys.version_info < (3, 9):
    # Python 3.9 does not include `_special`, so use the function from typing instead

    def get_args(tp):
        """Get type arguments with all substitutions performed.
        For unions, basic simplifications used by Union constructor are performed.
        Examples::
            get_args(Dict[str, int]) == (str, int)
            get_args(int) == ()
            get_args(Union[int, Union[T, int], str][int]) == (int, str)
            get_args(Union[int, Tuple[T, int]][str]) == (int, Tuple[str, int])
            get_args(Callable[[], T][int]) == ([], int)
        """
        if isinstance(tp, (_GenericAlias, _SpecialForm)) and not tp._special:
            res = tp.__args__
            if tp.__origin__ is collections.abc.Callable and res[0] is not Ellipsis:
                res = (list(res[:-1]), res[-1])
            return res
        return ()

else:
    from typing import get_args as get_args  # NOQA


def unpack_if_union(tp):
    if is_union(tp):
        return get_args(tp)[0]
    else:
        return tp


def is_union(tp):
    return isinstance(tp, _GenericAlias) and tp._name in (None, "Optional")


def is_list(tp):
    return isinstance(tp, _GenericAlias) and tp._name == "List"


def get_filtered_type_hints(class_type: Type) -> List[Tuple[str, Type]]:
    """Gets all the (valid) type hints for a given class.

    Args:
        class_type (Type): Class to extract property fields.

    Returns:
        List[Tuple[str, Type]]: Filtered list of tuples representing field name and type.
    """
    return [
        (field_name, field)
        for field_name, field in get_type_hints(class_type).items()
        if not field_name.startswith("__")
    ]


def get_required_class_field(class_type: Type) -> List[Tuple[str, Type]]:
    """Gets all the (valid) class fields which are mandatory.

    Args:
        class_type (Type): [description]

    Returns:
        List[Tuple[str, Type]]: [description]
    """
    return [
        (field_name, field)
        for field_name, field in class_type.__fields__.items()
        if field.required and not field_name.startswith("__")
    ]



import logging

from geolib.models.validators import BaseValidator

logger = logging.getLogger(__name__)


class DSheetPilingValidator(BaseValidator):
    """Validator for DSheetpiling internal structure.

    Raises:
        ValueError: When the internal datastructure is not valid including a message why.

    Has access to self.ds from parent class.
    Will run all is_valid_ functions to check for validity.
    """

    def is_valid_stages(self) -> bool:
        """
        When adding stages, some values are populated with None.
        These should be replaced before serializing.
        """
        for i, stage in enumerate(self.ds.construction_stages.stages):
            for key, value in dict(stage).items():
                if value is None:
                    raise ValueError(
                        f"All attributes must be populated; stage {i}-{stage.name} has None value for {key}"
                    )
                if "water_level" in key and not isinstance(value, str):
                    if value not in self.ds.waterlevels.water_level_names:
                        raise ValueError(
                            f"{key}-{value} in stage {i}-{stage.name} is not defined in internal data structure"
                        )
                    continue
                if "profile" in key and not isinstance(value, str):
                    if value not in self.ds.soil_profiles.soil_profile_names:
                        raise ValueError(
                            f"{key}-{value} in stage {i}-{stage.name} is not defined in internal data structure"
                        )
                    continue
                if "surface" in key and not isinstance(value, str):
                    if value not in self.ds.surfaces.surface_names:
                        raise ValueError(
                            f"{key}-{value} in stage {i}-{stage.name} is not defined in internal data structure"
                        )
                    continue
                if key == "anchors":
                    for anchor in value:
                        if anchor["name"] not in self.ds.anchors.anchor_names:
                            raise ValueError(
                                f"{key}-{value} in stage {i}-{stage.name} is not defined in internal data structure"
                            )
                    continue
                if key == "struts":
                    for strut in value:
                        if strut["name"] not in self.ds.struts.strut_names:
                            raise ValueError(
                                f"{key}-{value} in stage {i}-{stage.name} is not defined in internal data structure"
                            )
                    continue
        if len(self.ds.construction_stages.stages) != len(
            self.ds.calculation_options_per_stage.stageoptions
        ):
            error = f"Number of stages defined is {len(self.ds.construction_stages.stages)} which is not the same as {len(self.ds.calculation_options_per_stage.stageoptions)}"
            logger.warning(error)
            raise ValueError(error)

        return True

    def is_valid_soil_profiles(self) -> bool:
        """All soils in soil profiles should be defined in the soil collection"""
        if isinstance(self.ds.soil_profiles, str):
            return True  # Resultings from parsing input file.
        soil_names = self.ds.soil_collection.soil_names
        referenced_soil_names = self.ds.soil_profiles.referenced_soil_names
        if not referenced_soil_names.issubset(soil_names):
            raise ValueError(
                f"The following soils were not found in soil collection: {', '.join(referenced_soil_names - soil_names)}"
            )
        return True



from enum import IntEnum
from typing import Optional

from pydantic import PositiveFloat, confloat, constr

from geolib.models import BaseDataClass

from .internal import Anchor as InternalAnchor
from .internal import Strut as InternalStrut
from .internal import Support as InternalSupport
from .settings import Side


class Anchor(BaseDataClass):
    """Anchor. This option is not available for SinglePileModelType.

    Args:
        name: Name of the anchor
        level: Level of the anchor, or z-coordinate [m]
        e_modulus: E-modulus [kN/m^2]
        cross_section: Cross section [m^2/m']
        wall_height_kranz: Height of the wall [Kranz] [m]
        length: Length [m]
        angle: Angle [deg]
        yield_force: Yield force [kN/m']
        side: Side of the anchor [Side]
    """

    name: constr(min_length=1, max_length=50)
    level: float
    e_modulus: Optional[PositiveFloat] = None
    cross_section: Optional[PositiveFloat] = None
    wall_height_kranz: Optional[confloat(ge=0)] = None
    length: Optional[PositiveFloat] = None
    angle: Optional[float] = None
    side: Side = Side.RIGHT
    yield_force: Optional[confloat(ge=0)] = None

    def to_internal(self) -> InternalAnchor:
        return InternalAnchor(**self.dict(exclude_none=True))


class Strut(BaseDataClass):
    """Strut. This option is not available for SinglePileModelType.

    Args:
        name: Name of the strut
        level: Level of the strut, or z-coordinate [m]
        e_modulus: E-modulus [kN/m^2]
        cross_section: Cross section [m^2/m']
        length: Length [m]
        angle: Angle [deg]
        buckling_force: Buckling force of the strut [kN/m']
        side: Side of the strut [Side]
        pre_compression: Pre-compressions [kN/m']
    """

    name: constr(min_length=1, max_length=50)
    level: float
    e_modulus: Optional[PositiveFloat] = None
    cross_section: Optional[PositiveFloat] = None
    length: Optional[PositiveFloat] = None
    angle: Optional[float] = None
    buckling_force: Optional[confloat(ge=0)] = None
    side: Side = Side.RIGHT
    pre_compression: Optional[PositiveFloat] = None

    def to_internal(self) -> InternalStrut:
        return InternalStrut(**self.dict(exclude_none=True, exclude={"pre_compression"}))


class SupportType(IntEnum):
    TRANSLATION = 1
    ROTATION = 2
    TRANSLATION_AND_ROTATION = 3


class SpringSupport(BaseDataClass):
    """Spring support."""

    name: constr(min_length=1, max_length=50)
    level: float
    rotational_stiffness: confloat(ge=0)
    translational_stiffness: confloat(ge=0)

    def to_internal(self) -> InternalSupport:
        return InternalSupport(**self.dict())


class RigidSupport(BaseDataClass):
    """Rigid support."""

    name: constr(min_length=1, max_length=50)
    level: float
    support_type: SupportType

    def to_internal(self) -> InternalSupport:

        rotational_stiffness, translational_stiffness = {
            SupportType.TRANSLATION.value: (0, 1),
            SupportType.ROTATION.value: (1, 0),
            SupportType.TRANSLATION_AND_ROTATION.value: (1, 1),
        }[self.support_type.value]

        return InternalSupport(
            name=self.name,
            level=self.level,
            rotational_stiffness=rotational_stiffness,
            translational_stiffness=translational_stiffness,
        )



import abc
from math import isfinite

from pydantic import BaseModel

from .meta import MetaData
from .validators import BaseValidator

settings = MetaData()


class BaseDataClass(BaseModel):
    """Base class for *all* pydantic classes in GEOLib."""

    class Config:
        validate_assignment = True
        arbitrary_types_allowed = True
        validate_all = True
        extra = settings.extra_fields


class BaseModelStructure(BaseDataClass, abc.ABC):
    @property
    def is_valid(self) -> bool:
        """Validates the current model structure."""
        return self.validator().is_valid

    def validator(self) -> BaseValidator:
        """Set the Validator class."""
        return BaseValidator(self)



import logging
from typing import Set

from geolib.models.validators import BaseValidator

logger = logging.getLogger(__name__)


class DStabilityValidator(BaseValidator):
    """Validator for DStability structure.

    Has access to self.ds from parent class.
    Will run all is_valid_ functions to check for validity."""

    def is_valid_stages(self) -> bool:
        """Number of stages should be the same:"""
        lengths_set = set()
        valid = True
        for key, value in (
            (k, v) for k, v in self.ds.dict().items() if "result" not in k
        ):  # Results not required for stage validity.
            if isinstance(value, list):
                lengths_set.add(len(value))
                if len(lengths_set) > 1:
                    logger.error(
                        f"{self.is_valid_stages.__doc__} {key} has different number of stages: {len(value)}."
                    )
                    valid = False
                    break

        return valid

    def is_valid_layer_loads(self) -> bool:
        """Each layer load must have a consolidation degree for each soil layer"""
        for stage_id, _ in enumerate(self.ds.stages):
            soil_layer_ids: Set[str] = {
                layer.LayerId for layer in self.ds.soillayers[stage_id].SoilLayers
            }

            layer_load_layer_ids: Set[str] = set()
            for layer_load in self.ds.loads[stage_id].LayerLoads:
                layer_load_layer_ids.add(layer_load.LayerId)

                consolidation_layer_id_references: Set[str] = set()
                for consolidation in layer_load.Consolidations:
                    consolidation_layer_id_references.add(consolidation.LayerId)
                    if consolidation.LayerId is None:
                        return False
                if soil_layer_ids - consolidation_layer_id_references != set(
                    [layer_load.LayerId]
                ):  # All other soillayer ids are included except the soillayers own id.
                    return False
            if layer_load_layer_ids != soil_layer_ids:
                return False
        return True



class BaseValidator:
    def __init__(self, ds):
        self.ds = ds

    @property
    def is_valid(self) -> bool:
        return all(
            [
                getattr(self, func)()
                for func in dir(self)
                if (func.startswith("is_valid_"))
            ]
        )



from enum import Enum, IntEnum
from typing import Optional

from geolib.models.dseries_parser import DSeriesUnmappedNameProperties
from geolib.models.internal import Bool
from geolib.soils import DistributionType, HorizontalBehaviourType, StorageTypes


class PreconType(IntEnum):
    UNDEFINED = -1
    OCR = 0
    PRECONSOLIDATION_PRESSURE = 1
    POP = 2


class SoilInternal(DSeriesUnmappedNameProperties):
    """Representation of [SOIL] group."""

    name: str = ""
    soilcolor: int = 9764853
    soilgamdry: float = 14
    soilgamwet: float = 14
    soilinitialvoidratio: float = 0
    soilcohesion: float = 10  # fixed value cannot be modified
    soilphi: float = 30  # fixed value cannot be modified
    soilpreconisotachetype: PreconType = PreconType.POP
    soilpreconkoppejantype: PreconType = PreconType.OCR
    soiluseequivalentage: Bool = Bool.FALSE
    soilequivalentage: float = 19800
    soilpc: float = 0
    soilocr: float = 1
    soilpop: float = 0
    soillimitstress: float = 0  # fixed value
    soildrained: Bool = Bool.FALSE
    soilapasapproximationbycpcs: Bool = Bool.FALSE
    soilcv: float = 1e-12
    soilpermeabilityver: float = 0.00000001
    soilpermeabilityhorfactor: float = 1
    soilstoragetype: StorageTypes = StorageTypes.strain_dependent_permeability
    soilpermeabilitystrainmodulus: float = 1e15
    soiluseprobdefaults: Bool = Bool.FALSE
    soilstdgamdry: float = 0
    soilstdgamwet: float = 0
    soilstdcv: float = 0
    soilstdpc: float = 0
    soilstdpricompindex: float = 0.01  # SoilPriCompIndex
    soilstdseccompindex: float = 0.1  # SoilSecCompIndex
    soilstdseccomprate: float = 0.005  # SoilSecCompRate
    soilstdocr: float = 0
    soilstdpermeabilityver: float = 0
    soilstdpop: float = 0
    soilstdpermeabilityhorfactor: float = 0
    soilstdinitialvoidratio: float = 0
    soilstdpermeabilitystrainmodulus: float = 0  # fixed value
    soilstdlimitstress: float = 0  # fixed value
    soilstdcp: float = 0
    soilstdcp1: float = 0
    soilstdcs: float = 0
    soilstdcs1: float = 0
    soilstdap: float = 0
    soilstdasec: float = 0
    soilstdcar: float = 0
    soilstdca: float = 0
    soilstdrratio: float = 0
    soilstdcratio: float = 0
    soilstdsratio: float = 0  # fixed value
    soilstdcrindex: float = 0
    soilstdccindex: float = 0
    soilstdcswindex: float = 0  # fixed value
    soildistgamdry: DistributionType = DistributionType.Normal
    soildistgamwet: DistributionType = DistributionType.Normal
    soildistcv: DistributionType = DistributionType.Normal
    soildistdpc: DistributionType = DistributionType.Normal
    soildistpricompindex: DistributionType = DistributionType.Normal
    soildistseccompindex: DistributionType = DistributionType.Normal
    soildistseccomprate: DistributionType = DistributionType.Normal
    soildistocr: DistributionType = DistributionType.Normal
    soildistpermeabilityver: DistributionType = DistributionType.Normal
    soildistpop: DistributionType = DistributionType.Normal
    soildistpermeabilityhorfactor: DistributionType = DistributionType.Normal
    soildistinitialvoidratio: DistributionType = DistributionType.Normal
    soildistpermeabilitystrainmodulus: DistributionType = (
        DistributionType.Normal
    )  # fixed value
    soildistlimitstress: DistributionType = DistributionType.Normal  # fixed value
    soildistcp: DistributionType = DistributionType.Normal
    soildistcp1: DistributionType = DistributionType.Normal
    soildistcs: DistributionType = DistributionType.Normal
    soildistcs1: DistributionType = DistributionType.Normal
    soildistap: DistributionType = DistributionType.Normal
    soildistasec: DistributionType = DistributionType.Normal
    soildistcar: DistributionType = DistributionType.Normal
    soildistca: DistributionType = DistributionType.Normal
    soildistrratio: DistributionType = DistributionType.Normal
    soildistcratio: DistributionType = DistributionType.Normal
    soildistsratio: DistributionType = DistributionType.Normal  # fixed value
    soildistcrindex: DistributionType = DistributionType.Normal
    soildistccindex: DistributionType = DistributionType.Normal
    soildistcswindex: DistributionType = DistributionType.Normal  # fixed value
    soilcorcpcp1: float = 0
    soilcorcscp1: float = 0
    soilcorcs1cp1: float = 0
    soilcorapcp1: float = 0
    soilcoraseccp1: float = 0
    soilcorcrindexccindex: float = 0
    soilcorrratiocratio: float = 0
    soilcorcaccindexorcratio: float = 0
    soilcorpricompindexseccompindex: float = 0  # SoilPriCompIndex
    soilcorseccomprateseccompindex: float = 0  # SoilSecCompRate
    soilcp: float = 1
    soilcp1: float = 1
    soilcs: float = 1
    soilcs1: float = 1
    soilap: float = 1
    soilasec: float = 1
    soilcar: float = 10
    soilca: float = 1
    soilcompratio: Bool = Bool.TRUE
    soilrratio: float = 1
    soilcratio: float = 1
    soilsratio: float = 0  # fixed value
    soilcrindex: float = 1
    soilccindex: float = 1
    soilcswindex: float = 0  # fixed value
    soilpricompindex: float = 0.01  # SoilPriCompIndex
    soilseccompindex: float = 0.1  # SoilSecCompIndex
    soilseccomprate: float = 0.005  # SoilSecCompRate
    soilhorizontalbehaviourtype: HorizontalBehaviourType = HorizontalBehaviourType.Elastic
    soilelasticity: float = 1000
    soildefaultelasticity: Bool = Bool.TRUE



from abc import ABCMeta, abstractmethod
from datetime import timedelta
from typing import Optional

from pydantic import constr

from geolib.geometry import Point
from geolib.models import BaseDataClass

from .internal import (
    LoadValuesCircular,
    LoadValuesRectangular,
    LoadValuesTank,
    LoadValuesTrapeziform,
    LoadValuesUniform,
)
from .internal import OtherLoad as _OtherLoad
from .internal import TypeOtherLoads


class OtherLoad(BaseDataClass, metaclass=ABCMeta):
    """Other Load Class to inherit from."""

    load_type: Optional[TypeOtherLoads]

    @abstractmethod
    def _to_internal(self, time: timedelta, p: Point):
        """Creates internal datastructure from generic load"""


class TrapeziformLoad(OtherLoad):
    """
    Create a trapeziform load with the given name and properties.

    Arguments:
        gamma -- The weight of the load per m\ :sup:`3`

    .. image:: /figures/dsettlement/trapeziform.png
        :height: 200px
        :width: 400 px
        :scale: 50 %
        :align: center

    """

    gamma: float = 0
    height: float = 0
    xl: float = 0
    xm: float = 0
    xr: float = 0

    def _to_internal(self, time: timedelta, p: Point) -> _OtherLoad:

        load_values = LoadValuesTrapeziform(
            gamma=self.gamma,
            height=self.height,
            xl=self.xl,
            xm=self.xm,
            xr=self.xr,
            Xp=p.x,
            Yp=p.z,
        )
        other_load = _OtherLoad(
            time=time.days,
            load_type=TypeOtherLoads.Trapeziform,
            load_values_trapeziform=load_values,
        )
        return other_load


class CircularLoad(OtherLoad):
    """Create a circular load with the given name and properties.

    Arguments:
        weight: The mangitude of the load.
        alpha: The shape factor alpha is used to specify the shape of the contact pressure.

    .. image:: /figures/dsettlement/circular.png
        :height: 200px
        :width: 400 px
        :scale: 50 %
        :align: center

    """

    weight: float = 0
    alpha: float = 0
    R: float = 0.01

    def _to_internal(self, time: timedelta, p: Point) -> _OtherLoad:
        load_values = LoadValuesCircular(
            weight=self.weight,
            alpha=self.alpha,
            Xcp=p.x,
            Ycp=p.z,
            Zcp=p.y,
            R=self.R,
        )
        other_load = _OtherLoad(
            time=time.days,
            load_type=TypeOtherLoads.Circular,
            load_values_circular=load_values,
        )
        return other_load


class RectangularLoad(OtherLoad):
    """Create a rectangular load with the given name and properties

    Arguments:
        weight: The mangitude of the load.
        alpha: The shape factor alpha is used to specify the shape of the contact pressure

    .. image:: /figures/dsettlement/rectangural.png
        :height: 200px
        :width: 400 px
        :scale: 50 %
        :align: center

    """

    weight: float = 0  # kN/m2
    alpha: float = 0
    xwidth: float = 0.01
    zwidth: float = 0.01

    def _to_internal(self, time: timedelta, p: Point) -> _OtherLoad:
        load_values = LoadValuesRectangular(
            weight=self.weight,
            alpha=self.alpha,
            Xcp=p.x,
            Ycp=p.z,
            Zcp=p.y,
            xwidth=self.xwidth,
            zwidth=self.zwidth,
        )
        other_load = _OtherLoad(
            time=time.days,
            load_type=TypeOtherLoads.Rectangular,
            load_values_rectangular=load_values,
        )
        return other_load


# TODO This is a 1D Load and should not be other load?
class UniformLoad(OtherLoad):
    """Create a uniform load with the given name and properties.

    Arguments:
        unit_weight: The weight of the load per m\ :sup:`3`

    .. image:: /figures/dsettlement/uniform.png
        :height: 200px
        :width: 400 px
        :scale: 50 %
        :align: center

    """

    unit_weight: float = 0
    gamma: float = 0
    height: float = 0
    y_application: float = 0

    def _to_internal(self, time: timedelta, point: Point) -> _OtherLoad:
        load_values = LoadValuesUniform(
            unit_weight=self.unit_weight,
            height=self.height,
            y_application=point.z,
            gamma=self.gamma,
        )
        other_load = _OtherLoad(
            time=time.days,
            load_type=TypeOtherLoads.Uniform,
            load_values_uniform=load_values,
        )
        return other_load


class TankLoad(OtherLoad):
    """Create a tank load with the given name and properties.

    Arguments:
        wallweight: The magnitude of the load induced by the weight of the material in which the tank is made.
        alpha: The shape factor alpha is used to specify the shape of the contact pressure.
        internalweight: The magnitude of the load induced by the weight of the material stored in the tank.

    .. image:: /figures/dsettlement/tank.png
        :height: 200px
        :width: 400 px
        :scale: 50 %
        :align: center

    """

    wallweight: float = 0
    internalweight: float = 0
    alpha: float = 0
    Rintern: float = 0.01
    dWall: float = 0.01

    def _to_internal(self, time: timedelta, p: Point) -> _OtherLoad:
        load_values = LoadValuesTank(
            wallweight=self.wallweight,
            alpha=self.alpha,
            internalweight=self.internalweight,
            Xcp=p.x,
            Ycp=p.z,
            Zcp=p.y,
            Rintern=self.Rintern,
            dWall=self.dWall,
        )
        other_load = _OtherLoad(
            time=time.days,
            load_type=TypeOtherLoads.Tank,
            load_values_tank=load_values,
        )
        return other_load



"""Pile Library for D-Foundations.

"""
from enum import Enum
from typing import Optional

from pydantic.types import PositiveInt, confloat, constr

from geolib.geometry import Point
from geolib.models import BaseDataClass

from .internal import (
    BearingPileSlipLayer,
    LoadSettlementCurve,
    PileMaterial,
    PileShape,
    PileType,
    PileTypeForClayLoamPeat,
    PositionBearingPile,
    PositionTensionPile,
    TypesBearingPiles,
    TypesTensionPiles,
)


class PileLocation(BaseDataClass):
    """Base Class for Pile location."""

    pile_name: constr(min_length=0, max_length=10) = ""
    point: Point
    pile_head_level: confloat(ge=-1000, le=1000)
    limit_state_str: confloat(ge=0, le=100000)
    limit_state_service: confloat(ge=0, le=100000)

    def _to_internal(self, index: PositiveInt):
        return {
            "index": index,
            "pile_name": "'" + self.pile_name + "'",
            "x_coordinate": self.point.x,
            "y_coordinate": self.point.y,
            "pile_head_level": self.pile_head_level,
            "limit_state_str": self.limit_state_str,
            "limit_state_service": self.limit_state_service,
        }


class BearingPileLocation(PileLocation):
    """Inherits :class:`~geolib.models.dfoundations.piles.PileLocation`."""

    surcharge: confloat(ge=0, le=100000)

    def _to_internal(self, index: PositiveInt):
        pile_location = super()._to_internal(index)
        return PositionBearingPile(**pile_location, surcharge=self.surcharge)


class TensionPileLocation(PileLocation):
    """Inherits :class:`~geolib.models.dfoundations.piles.PileLocation`."""

    use_alternating_loads: bool
    max_force: confloat(ge=-100000, le=100000)
    min_force: confloat(ge=-100000, le=100000)

    def _to_internal(self, index: PositiveInt):
        pile_location = super()._to_internal(index)
        return PositionTensionPile(
            **pile_location,
            use_alternating_loads=self.use_alternating_loads,
            max_force=self.max_force,
            min_force=self.min_force,
        )


class BasePileType(Enum):
    """Supported pile types enum"""

    USER_DEFINED_VIBRATING = PileType.USER_DEFINED_VIBRATING.value
    USER_DEFINED_LOW_VIBRATING = PileType.USER_DEFINED_LOW_VIBRATING.value


class BasePileTypeForClayLoamPeat(Enum):
    """Pile types for clay loam and peat enum"""

    STANDARD = PileTypeForClayLoamPeat.STANDARD.value
    USER_DEFINED = PileTypeForClayLoamPeat.USER_DEFINED.value


class Pile(BaseDataClass):
    """Base Class for Piles."""

    pile_name: str
    pile_type: BasePileType
    pile_class_factor_shaft_sand_gravel: confloat(ge=0, le=9)
    pile_class_factor_shaft_clay_loam_peat: Optional[confloat(ge=0, le=9)]
    preset_pile_class_factor_shaft_clay_loam_peat: BasePileTypeForClayLoamPeat
    elasticity_modulus: confloat(ge=0, le=1e25)


class BearingPile(Pile):
    """Inherits :class:`~geolib.models.dfoundations.piles.Pile`."""

    pile_class_factor_tip: confloat(ge=0, le=9)
    load_settlement_curve: LoadSettlementCurve
    user_defined_pile_type_as_prefab: bool
    use_manual_reduction_for_qc: bool
    reduction_percentage_qc: confloat(ge=25, le=100) = 25
    characteristic_adhesion: confloat(ge=0, le=1000)

    overrule_pile_tip_shape_factor: bool
    pile_tip_shape_factor: Optional[confloat(ge=0, le=10)]
    overrule_pile_tip_cross_section_factors: bool
    pile_tip_cross_section_factor: Optional[confloat(ge=0, le=10)]

    def _to_internal(self):
        return TypesBearingPiles(
            pile_name=self.pile_name,
            pile_type=self.pile_type.value,
            pile_type_for_execution_factor_sand_gravel=PileType.USER_DEFINED,
            execution_factor_sand_gravel=self.pile_class_factor_shaft_sand_gravel,
            pile_type_for_execution_factor_clay_loam_peat=(
                self.preset_pile_class_factor_shaft_clay_loam_peat.value
            ),
            execution_factor_clay_loam_peat=self.pile_class_factor_shaft_clay_loam_peat,
            pile_type_for_pile_class_factor=PileType.USER_DEFINED,
            pile_class_factor=self.pile_class_factor_tip,
            pile_type_for_load_settlement_curve=self.load_settlement_curve,
            user_defined_pile_type_as_prefab=self.user_defined_pile_type_as_prefab,
            use_manual_reduction_for_qc=self.use_manual_reduction_for_qc,
            reduction_percentage_qc=self.reduction_percentage_qc,
            material=PileMaterial.USER_DEFINED,
            elasticity_modulus=self.elasticity_modulus,
            slip_layer=BearingPileSlipLayer.USER_DEFINED,
            characteristic_adhesion=self.characteristic_adhesion,
            overrule_pile_tip_shape_factor=self.overrule_pile_tip_shape_factor,
            pile_tip_shape_factor=self.pile_tip_shape_factor,
            overrule_pile_tip_cross_section_factors=(
                self.overrule_pile_tip_cross_section_factors
            ),
            pile_tip_cross_section_factor=self.pile_tip_cross_section_factor,
            is_user_defined=True,
            use_pre_2016=False,
        )


class TensionPile(Pile):
    """Inherits :class:`~geolib.models.dfoundations.piles.Pile`."""

    unit_weight_pile: confloat(ge=0, le=1000)

    def _to_internal(self):
        return TypesTensionPiles(
            pile_name=self.pile_name,
            pile_type_for_execution_factor_sand_gravel=self.pile_type.value,
            execution_factor_sand_gravel=self.pile_class_factor_shaft_sand_gravel,
            pile_type_for_execution_factor_clay_loam_peat=(
                self.preset_pile_class_factor_shaft_clay_loam_peat.value
            ),
            execution_factor_clay_loam_peat=self.pile_class_factor_shaft_clay_loam_peat,
            material=PileMaterial.USER_DEFINED,
            unit_weight_pile=self.unit_weight_pile,
            elasticity_modulus=self.elasticity_modulus,
        )


class BearingRoundPile(BearingPile):
    """Inherits :class:`~geolib.models.dfoundations.piles.BearingPile`."""

    diameter: confloat(ge=0, le=100)

    def _to_internal(self):
        types_bearing_pile = super()._to_internal()
        types_bearing_pile.diameter = self.diameter
        types_bearing_pile.shape = PileShape.ROUND_PILE
        return types_bearing_pile


class TensionRoundPile(TensionPile):
    """Inherits :class:`~geolib.models.dfoundations.piles.TensionPile`."""

    diameter: confloat(ge=0, le=100)

    def _to_internal(self):
        types_tension_pile = super()._to_internal()
        types_tension_pile.diameter = self.diameter
        types_tension_pile.shape = PileShape.ROUND_PILE
        return types_tension_pile


class BearingRectangularPile(BearingPile):
    """Inherits :class:`~geolib.models.dfoundations.piles.BearingPile`."""

    base_width: confloat(ge=0, le=100)
    base_length: confloat(ge=0, le=100)

    def _to_internal(self):
        types_bearing_pile = super()._to_internal()
        types_bearing_pile.base_width = self.base_width
        types_bearing_pile.base_length = self.base_length
        types_bearing_pile.shape = PileShape.RECTANGULAR_PILE
        return types_bearing_pile


class TensionRectangularPile(TensionPile):
    """Inherits :class:`~geolib.models.dfoundations.piles.TensionPile`."""

    base_width: confloat(ge=0, le=100)
    base_length: confloat(ge=0, le=100)

    def _to_internal(self):
        types_tension_pile = super()._to_internal()
        types_tension_pile.base_width = self.base_width
        types_tension_pile.base_length = self.base_length
        types_tension_pile.shape = PileShape.RECTANGULAR_PILE
        return types_tension_pile


class BearingRoundPileWithEnlargedBase(BearingPile):
    """Inherits :class:`~geolib.models.dfoundations.piles.BearingPile`."""

    base_diameter: confloat(ge=0, le=100)
    pile_diameter: confloat(ge=0, le=100)
    base_height: confloat(ge=0, le=100)

    def _to_internal(self):
        types_bearing_pile = super()._to_internal()
        types_bearing_pile.base_diameter = self.base_diameter
        types_bearing_pile.pile_diameter = self.pile_diameter
        types_bearing_pile.base_height = self.base_height
        types_bearing_pile.shape = PileShape.ROUND_PILE_WITH_ENLARGED_BASE
        return types_bearing_pile


class TensionRoundPileWithEnlargedBase(TensionPile):
    """Inherits :class:`~geolib.models.dfoundations.piles.TensionPile`."""

    base_diameter: confloat(ge=0, le=100)
    pile_diameter: confloat(ge=0, le=100)
    base_height: confloat(ge=0, le=100)

    def _to_internal(self):
        types_tension_pile = super()._to_internal()
        types_tension_pile.base_diameter = self.base_diameter
        types_tension_pile.pile_diameter = self.pile_diameter
        types_tension_pile.base_height = self.base_height
        types_tension_pile.shape = PileShape.ROUND_PILE_WITH_ENLARGED_BASE
        return types_tension_pile


class BearingRectangularPileWithEnlargedBase(BearingPile):
    """Inherits :class:`~geolib.models.dfoundations.piles.BearingPile`."""

    base_width_v: confloat(ge=0, le=100)
    base_length_v: confloat(ge=0, le=100)
    base_height: confloat(ge=0, le=100)
    shaft_width: confloat(ge=0, le=100)
    shaft_length: confloat(ge=0, le=100)

    def _to_internal(self):
        types_bearing_pile = super()._to_internal()
        types_bearing_pile.base_width_v = self.base_width_v
        types_bearing_pile.base_length_v = self.base_length_v
        types_bearing_pile.base_height = self.base_height
        types_bearing_pile.shaft_width = self.shaft_width
        types_bearing_pile.shaft_length = self.shaft_length
        types_bearing_pile.shape = PileShape.RECTANGULAR_PILE_WITH_ENLARGED_BASE
        return types_bearing_pile


class TensionRectangularPileWithEnlargedBase(TensionPile):
    """Inherits :class:`~geolib.models.dfoundations.piles.TensionPile`."""

    base_width_v: confloat(ge=0, le=100)
    base_length_v: confloat(ge=0, le=100)
    base_height: confloat(ge=0, le=100)
    shaft_width: confloat(ge=0, le=100)
    shaft_length: confloat(ge=0, le=100)

    def _to_internal(self):
        types_tension_pile = super()._to_internal()
        types_tension_pile.base_width_v = self.base_width_v
        types_tension_pile.base_length_v = self.base_length_v
        types_tension_pile.base_height = self.base_height
        types_tension_pile.shaft_width = self.shaft_width
        types_tension_pile.shaft_length = self.shaft_length
        types_tension_pile.shape = PileShape.RECTANGULAR_PILE_WITH_ENLARGED_BASE
        return types_tension_pile


class BearingRoundTaperedPile(BearingPile):
    """Inherits :class:`~geolib.models.dfoundations.piles.BearingPile`."""

    diameter_at_pile_tip: confloat(ge=0, le=100)
    increase_in_diameter: confloat(ge=0, le=100)

    def _to_internal(self):
        types_bearing_pile = super()._to_internal()
        types_bearing_pile.diameter = self.diameter_at_pile_tip
        types_bearing_pile.increase_in_diameter = self.increase_in_diameter
        types_bearing_pile.shape = PileShape.ROUND_TAPERED_PILE
        return types_bearing_pile


class TensionRoundTaperedPile(TensionPile):
    """Inherits :class:`~geolib.models.dfoundations.piles.TensionPile`."""

    diameter_at_pile_tip: confloat(ge=0, le=100)
    increase_in_diameter: confloat(ge=0, le=100)

    def _to_internal(self):
        types_tension_pile = super()._to_internal()
        types_tension_pile.diameter = self.diameter_at_pile_tip
        types_tension_pile.increase_in_diameter = self.increase_in_diameter
        types_tension_pile.shape = PileShape.ROUND_TAPERED_PILE
        return types_tension_pile


class BearingRoundHollowPileWithClosedBase(BearingPile):
    """Inherits :class:`~geolib.models.dfoundations.piles.BearingPile`."""

    external_diameter: confloat(ge=0, le=100)
    wall_thickness: confloat(ge=0, le=100)

    def _to_internal(self):
        types_bearing_pile = super()._to_internal()
        types_bearing_pile.external_diameter = self.external_diameter
        types_bearing_pile.internal_diameter = (
            self.external_diameter - 2 * self.wall_thickness
        )
        types_bearing_pile.shape = PileShape.ROUND_HOLLOW_PILE_WITH_CLOSED_BASE
        return types_bearing_pile


class TensionRoundHollowPileWithClosedBase(TensionPile):
    """Inherits :class:`~geolib.models.dfoundations.piles.TensionPile`."""

    external_diameter: confloat(ge=0, le=100)
    wall_thickness: confloat(ge=0, le=100)

    def _to_internal(self):
        types_tension_pile = super()._to_internal()
        types_tension_pile.external_diameter = self.external_diameter
        types_tension_pile.internal_diameter = (
            self.external_diameter - 2 * self.wall_thickness
        )
        types_tension_pile.shape = PileShape.ROUND_HOLLOW_PILE_WITH_CLOSED_BASE
        return types_tension_pile


class BearingRoundPileWithLostTip(BearingPile):
    """Inherits :class:`~geolib.models.dfoundations.piles.BearingPile`."""

    base_diameter: confloat(ge=0, le=100)
    pile_diameter: confloat(ge=0, le=100)

    def _to_internal(self):
        types_bearing_pile = super()._to_internal()
        types_bearing_pile.base_diameter = self.base_diameter
        types_bearing_pile.pile_diameter = self.pile_diameter
        types_bearing_pile.shape = PileShape.ROUND_PILE_WITH_LOST_TIP
        return types_bearing_pile


class TensionRoundPileWithLostTip(TensionPile):
    """Inherits :class:`~geolib.models.dfoundations.piles.TensionPile`."""

    base_diameter: confloat(ge=0, le=100)
    pile_diameter: confloat(ge=0, le=100)

    def _to_internal(self):
        types_tension_pile = super()._to_internal()
        types_tension_pile.base_diameter = self.base_diameter
        types_tension_pile.pile_diameter = self.pile_diameter
        types_tension_pile.shape = PileShape.ROUND_PILE_WITH_LOST_TIP
        return types_tension_pile


class BearingRoundPileWithInSituFormedBase(BearingPile):
    """Inherits :class:`~geolib.models.dfoundations.piles.BearingPile`."""

    base_diameter: confloat(ge=0, le=100)
    pile_diameter: confloat(ge=0, le=100)
    base_height: confloat(ge=0, le=100)

    def _to_internal(self):
        types_bearing_pile = super()._to_internal()
        types_bearing_pile.base_diameter = self.base_diameter
        types_bearing_pile.pile_diameter = self.pile_diameter
        types_bearing_pile.base_height = self.base_height
        types_bearing_pile.shape = PileShape.ROUND_PILE_WITH_IN_SITU_FORMED_BASE
        return types_bearing_pile


class TensionRoundPileWithInSituFormedBase(TensionPile):
    """Inherits :class:`~geolib.models.dfoundations.piles.TensionPile`."""

    base_diameter: confloat(ge=0, le=100)
    pile_diameter: confloat(ge=0, le=100)
    base_height: confloat(ge=0, le=100)

    def _to_internal(self):
        types_tension_pile = super()._to_internal()
        types_tension_pile.base_diameter = self.base_diameter
        types_tension_pile.pile_diameter = self.pile_diameter
        types_tension_pile.base_height = self.base_height
        types_tension_pile.shape = PileShape.ROUND_PILE_WITH_IN_SITU_FORMED_BASE
        return types_tension_pile


class BearingSection(BearingPile):
    """Inherits :class:`~geolib.models.dfoundations.piles.BearingPile`."""

    base_width: confloat(ge=0, le=100)
    base_length: confloat(ge=0, le=100)

    def _to_internal(self):
        types_bearing_pile = super()._to_internal()
        types_bearing_pile.base_width = self.base_width
        types_bearing_pile.base_length = self.base_length
        types_bearing_pile.shape = PileShape.SECTION
        return types_bearing_pile


class TensionSection(TensionPile):
    """Inherits :class:`~geolib.models.dfoundations.piles.TensionPile`."""

    circumference: confloat(ge=0, le=100)
    cross_section: confloat(ge=0, le=100)

    def _to_internal(self):
        types_tension_pile = super()._to_internal()
        types_tension_pile.circumference = self.circumference
        types_tension_pile.cross_section = self.cross_section
        types_tension_pile.shape = PileShape.USER_DEFINED
        return types_tension_pile


class BearingRoundOpenEndedHollowPile(BearingPile):
    """Inherits :class:`~geolib.models.dfoundations.piles.BearingPile`."""

    external_diameter: confloat(ge=0, le=100)
    wall_thickness: confloat(ge=0, le=100)

    def _to_internal(self):
        types_bearing_pile = super()._to_internal()
        types_bearing_pile.external_diameter = self.external_diameter
        types_bearing_pile.internal_diameter = (
            self.external_diameter - 2 * self.wall_thickness
        )
        types_bearing_pile.shape = PileShape.ROUND_OPEN_ENDED_HOLLOW_PILE
        return types_bearing_pile


class TensionRoundOpenEndedHollowPile(TensionPile):
    """Inherits :class:`~geolib.models.dfoundations.piles.TensionPile`."""

    external_diameter: confloat(ge=0, le=100)
    wall_thickness: confloat(ge=0, le=100)

    def _to_internal(self):
        types_tension_pile = super()._to_internal()
        types_tension_pile.external_diameter = self.external_diameter
        types_tension_pile.internal_diameter = (
            self.external_diameter - 2 * self.wall_thickness
        )
        types_tension_pile.shape = PileShape.ROUND_OPEN_ENDED_HOLLOW_PILE
        return types_tension_pile


class BearingHShapedPile(BearingPile):
    """Inherits :class:`~geolib.models.dfoundations.piles.BearingPile`."""

    height_h_shape: confloat(ge=0, le=100)
    width_h_shape: confloat(ge=0, le=100)
    thickness_web: confloat(ge=0, le=100)
    thickness_flange: confloat(ge=0, le=100)

    def _to_internal(self):
        types_bearing_pile = super()._to_internal()
        types_bearing_pile.height_h_shape = self.height_h_shape
        types_bearing_pile.width_h_shape = self.width_h_shape
        types_bearing_pile.thickness_web = self.thickness_web
        types_bearing_pile.thickness_flange = self.thickness_flange
        types_bearing_pile.shape = PileShape.H_SHAPED_PROFILE
        return types_bearing_pile


class TensionHShapedPile(TensionPile):
    """Inherits :class:`~geolib.models.dfoundations.piles.TensionPile`."""

    height_h_shape: confloat(ge=0, le=100)
    width_h_shape: confloat(ge=0, le=100)
    thickness_web: confloat(ge=0, le=100)
    thickness_flange: confloat(ge=0, le=100)

    def _to_internal(self):
        types_tension_pile = super()._to_internal()
        types_tension_pile.height_h_shape = self.height_h_shape
        types_tension_pile.width_h_shape = self.width_h_shape
        types_tension_pile.thickness_web = self.thickness_web
        types_tension_pile.thickness_flange = self.thickness_flange
        types_tension_pile.shape = PileShape.H_SHAPED_PROFILE
        return types_tension_pile



import logging
from datetime import timedelta
from operator import attrgetter
from pathlib import Path
from subprocess import CompletedProcess, run
from typing import BinaryIO
from typing import List, Optional, Type, Union

from pydantic import FilePath, validate_arguments
from pydantic.types import PositiveInt, confloat, conint, constr

from geolib.geometry import Point
from geolib.models import BaseDataClass, BaseModel, BaseModelStructure
from geolib.models.dsettlement.internal_soil import SoilInternal
from geolib.models.dsettlement.loads import (
    CircularLoad,
    RectangularLoad,
    TankLoad,
    TrapeziformLoad,
    UniformLoad,
)
from geolib.models.dsettlement.probabilistic_calculation_types import (
    ProbabilisticCalculationType,
)
from geolib.models.dsettlement.serializer import DSettlementInputSerializer
from geolib.models.meta import CONSOLE_RUN_BATCH_FLAG, MetaData
from geolib.soils import DistributionType
from geolib.soils import Soil as Soil_Input

from .drains import VerticalDrain
from .dsettlement_parserprovider import DSettlementParserProvider
from .internal import (
    Boundary,
    CalculationOptions,
    ConsolidationModel,
    Curve,
    Dimension,
    DSeriePoint,
    DSettlementOutputStructure,
    DSettlementStructure,
    Layer,
    Layers,
    Model,
    NonUniformLoad,
    NonUniformLoads,
    OtherLoads,
    PiezoLines,
    PointForLoad,
    Points,
    ResidualTimes,
    Results,
    SoilModel,
    StrainType,
    Verticals,
    WaterLoads,
)

logger = logging.getLogger(__name__)


class DSettlementModel(BaseModel):
    r"""
    D-Settlement is a dedicated tool for predicting soil settlements
    by external loading.

    This model can read, modify and create
    \*.sli files, read \*.sld and \*.err files.
    """

    datastructure: Union[
        DSettlementStructure, DSettlementOutputStructure
    ] = DSettlementStructure()

    @property
    def parser_provider_type(self) -> Type[DSettlementParserProvider]:
        return DSettlementParserProvider

    @property
    def console_path(self) -> Path:
        return Path("DSettlementConsole/DSettlementConsole.exe")

    @property
    def console_flags(self) -> List[str]:
        return [CONSOLE_RUN_BATCH_FLAG]

    def serialize(self, filename: Union[FilePath, BinaryIO]):
        """
        Serialize and pre-process
        Args:
            filename:
        """
        self.datastructure.geometry_data.pre_process()

        serializer = DSettlementInputSerializer(ds=self.datastructure.dict())
        serializer.write(filename)

        if isinstance(filename, Path):
            self.filename = filename

    def add_soil(self, soil_input: Soil_Input) -> None:
        """Soil is converted in the internal structure and added in soil_collection."""
        soil_new = soil_input._to_dsettlement()
        self.datastructure.soil_collection.add_soil_if_unique(soil_new)

    # 1.2.3 Models
    def set_model(
        self,
        constitutive_model: SoilModel,
        consolidation_model: ConsolidationModel,
        is_two_dimensional: bool,
        strain_type: StrainType,
        is_vertical_drain: bool,
        is_fit_for_settlement_plate: bool,
        is_probabilistic: bool,
        is_horizontal_displacements: bool,
        is_secondary_swelling: bool,
        is_waspan: bool,
    ):
        """
        Sets the D-settlement Model. Initializes CalculationOptions and Model if the type is str

        Args:
            constitutive_model (SoilModel): enum
            consolidation_model (ConsolidationModel): enum
            is_two_dimensional (bool): true if geometry is 2 dimensional
            strain_type (StrainType): enum
            is_vertical_drain (bool): true if vertical drain is present
            is_fit_for_settlement_plate (bool): true if fit for settlement plate
            is_probabilistic (bool): true if probabilistic calculation should be made
            is_horizontal_displacements (bool): true if horizontal displacements should be calculated
            is_secondary_swelling (bool): true if secondary swelling is present
            is_waspan (bool): true if waspan

        Returns:
            Model
        """
        if isinstance(self.datastructure.calculation_options, str):
            logger.warning("Calculation options are overwritten")
            self.datastructure.calculation_options = CalculationOptions()

        if isinstance(self.datastructure.model, str):
            logger.warning("Model options are overwritten")
            self.datastructure.model = Model()

        model_options = self.datastructure.model

        model_options.dimension = (
            Dimension.TWO_D if is_two_dimensional else Dimension.ONE_D
        )
        model_options.consolidation_model = consolidation_model
        model_options.soil_model = constitutive_model
        model_options.strain_type = strain_type
        model_options.is_vertical_drains = is_vertical_drain
        model_options.is_fit_for_settlement_plate = is_fit_for_settlement_plate
        model_options.is_probabilistic = is_probabilistic
        model_options.is_horizontal_displacements = is_horizontal_displacements
        model_options.is_secondary_swelling = is_secondary_swelling
        model_options.is_waspan = is_waspan

        return model_options

    def set_any_calculation_options(self, **kwargs):
        """
        Sets calculation options and initializes or removes data when necessary

        Args:
            precon_pressure_within_layer (PreconPressureWithinLayer): type of preconsolidation pressure within the layer
            is_imaginary_surface (bool): true if imaginary surface
            imaginary_surface_layer (PositivInt): index of layer
            is_submerging (bool): true if submerging
            use_end_time_for_fit (bool): true if end time should be used for fit
            is_maintain_profile (bool): true if profile should be maintained
            maintain_profile_material_name (str): name of the profile
            maintain_profile_time (conint(ge=0, le=100000)): time for maintain profile [days]
            maintain_profile_gamma_dry (confloat(ge=-100, le=100)): unit weight above phreatic line for maintain profile [kN/m3]
            maintain_profile_gamma_wet (confloat(ge=-100, le=100)): unit weight below phreatic line for maintain profile [kN/m3]
            dispersion_conditions_layer_boundaries_top (DispersionConditionLayerBoundary): dispersion condition at the top of the layer
            dispersion_conditions_layer_boundaries_top (DispersionConditionLayerBoundary): dispersion condition at the bottom of the layer
            stress_distribution_soil (StressDistributionSoil): type of stress distribution of the soil
            stress_distribution_loads (StressDistributionLoads): type of stress distribution loads
            iteration_stop_criteria_submerging (confloat(ge=0.0, le=1.0)): submerging iteration stop criteria
            iteration_stop_criteria_submerging_layer_height(confloat(ge=0, le=99.999)):  minimum settlement for submerging iteration stop criteria [m]
            maximum_iteration_steps_for_submerging (conint(ge=1, le=100)): maximum iteraion steps for submerging
            iteration_stop_criteria_desired_profile (confloat(ge=0, le=1)): iteration stop criteria for desired profile
            load_column_width_imaginary_surface (confloat(ge=0.05, le=10000)): load column width of the imaginary surface [m]
            load_column_width_non_uniform_loads (confloat(ge=0.05, le=10000)): load column width of non-uniform loads [m]
            load_column_width_trapeziform_loads (confloat(ge=0.05, le=10000)): load column width of trapeziform loads [m]
            end_of_consolidation (conint(ge=1, le=100000)): end of settlement calculation [days]
            number_of_subtime_steps (conint(ge=1, le=100)): number of subtime steps [-]
            reference_time (confloat(ge=0.001, le=1000000)): reference time [day]
            dissipation (bool): true if dissipation calculation should be added
            x_coord_dissipation (float): x-coordinate of the dissipation vertical [m]
            use_fit_factors (bool): true if fit parameters should be used
            x_coord_fit (float): x-coordinate of the fit [m]
            is_predict_settlements_omitting_additional_load_steps (bool): true if output of settlements by partial loading

        Returns:
            calculation options
        """
        if isinstance(self.datastructure.calculation_options, str):
            logger.warning("Calculation options are overwritten")
            self.datastructure.calculation_options = CalculationOptions()

        if isinstance(self.datastructure.model, str):
            logger.warning("Model options are overwritten")
            self.datastructure.model = Model()

        calculation_options = self.datastructure.calculation_options.dict()
        calculation_options.update(**kwargs)
        self.datastructure.calculation_options = CalculationOptions.set_options(
            **calculation_options
        )

        return calculation_options

    @property
    def accuracy(self):
        return self.datastructure.geometry_data.accuracy

    @property
    def curves(self):
        return self.datastructure.geometry_data.curves

    @property
    def boundaries(self):
        return self.datastructure.geometry_data.boundaries

    @property
    def use_probabilistic_defaults_boundaries(self):
        return self.datastructure.geometry_data.use_probabilistic_defaults_boundaries

    @property
    def stdv_boundaries(self):
        return self.datastructure.geometry_data.stdv_boundaries

    @property
    def distribution_boundaries(self):
        return self.datastructure.geometry_data.distribution_boundaries

    def set_probabilistic_data(
        self,
        point_of_vertical: Point,
        residual_settlement: float,
        maximum_number_of_samples: int,
        maximum_iterations: int,
        reliability_type: ProbabilisticCalculationType,
        is_reliability_calculation: bool,
    ):
        """
        Set calculation options for probabilistic calculation.

        Args:
            point_of_vertical : Select point that correspond to a defined vertical for the reliability analysis.
            residual_settlement : For FORM and Monte Carlo methods, the allowed residual settlement can be set.
            maximum_number_of_samples : The number of samples that the Monte Carlo method will use.
            maximum_iterations : The maximum number of iterations for the FORM method.
            reliability_type : Select one of the following methods

                - SettlementsDeterministic: a regular deterministic settlement analysis along all verticals, based on fixed mean values of the parameters.
                - BandWidthOfSettlementsFOSM: Quick and approximate determination of the bandwidth and the influencing factors (parameter
                    sensitivity) for the total settlements along one vertical. The determination is executed at user defined time points and at the time points
                    of measurements. Calculation time will increase with an increasing number of stochastic parameters.
                - ProbabilityOfFailureFORM: Iterative determination of the reliability index, bandwidth and influencing factors for the residual
                    settlement along one vertical. A separate FORM analysis is performed for each residual settlement that starts from each different
                    user defined time point. Calculation time will increase with an increasing number of stochastic parameters, user defined time points
                    and iterations. Furthermore, the FORM method is only conditionally stable.
                - BandWidthAndProbabilityOfFailureMonteCarlo: Determination of the bandwidth for the total settlements along one vertical, and also of the reliability index
                    and bandwidth for the residual settlements, by repetitive execution of settlement analyses (sampling). Each sample is executed with
                    random parameter values, derived from the stochastic distributions. Calculation time will increase with the number of samples. Accurate
                    Monte Carlo analysis requires a large number of samples, if many stochastic parameters are involved.

            is_reliability_calculation : set to True if a probabilistic calculation should be performed.

        """
        self.datastructure.check_x_in_vertical(point_of_vertical=point_of_vertical)
        self.datastructure.probabilistic_data = (
            self.datastructure.probabilistic_data.set_probabilistic_data(
                point_of_vertical=point_of_vertical,
                residual_settlement=residual_settlement,
                maximum_number_of_samples=maximum_number_of_samples,
                maximum_iterations=maximum_iterations,
                reliability_type=reliability_type,
                is_reliability_calculation=is_reliability_calculation,
            )
        )

    # 1.2.1 Soil profile
    # To create multiple layers
    def add_boundary(
        self,
        points: List[Point],
        use_probabilistic_defaults=True,
        stdv=0,
        distribution_boundaries=DistributionType.Undefined,
    ) -> int:
        """Add boundary to model."""
        # Divide points into curves and boundary
        # Check point uniqueness
        tolerance = self.accuracy.accuracy
        points = [
            self.points.add_point_if_unique(
                DSeriePoint.from_point(point), tolerance=tolerance
            )
            for point in points
        ]
        sorted_points = sorted(points, key=lambda point: (point.X, point.Y, point.Z))
        curves = self.curves.create_curves(sorted_points)
        boundary = self.boundaries.create_boundary(curves)
        # Probabilistic input for boundaries
        self.use_probabilistic_defaults_boundaries.append_use_probabilistic_defaults_boundary(
            use_probabilistic_defaults
        )
        self.stdv_boundaries.append_stdv_boundary(stdv)
        self.distribution_boundaries.append_distribution_boundary(distribution_boundaries)
        return boundary.id

    @property
    def points(self):
        """Enables easy access to the points in the internal dict-like datastructure. Also enables edit/delete for individual points."""
        return self.datastructure.geometry_data.points

    @property
    def headlines(self):
        return self.datastructure.geometry_data.piezo_lines

    def add_head_line(self, points: List[Point], is_phreatic=False) -> int:
        """Add head line to model."""
        # New points have to be created, whether points at the same location exist or not
        # This means that new points from geometry will re-use head line points,
        # So first create geometry, then add headlines.
        points = [
            self.points.add_point(DSeriePoint.from_point(point)) for point in points
        ]
        sorted_points = sorted(points, key=lambda point: (point.X, point.Y, point.Z))
        curves = self.curves.create_curves(sorted_points)

        piezo_line = self.headlines.create_piezoline(curves)
        if is_phreatic:
            self.datastructure.geometry_data.phreatic_line.phreatic_line = piezo_line.id
        return piezo_line.id

    @property
    def layers(self):
        return self.datastructure.geometry_data.layers

    def add_layer(
        self,
        boundary_top: int,
        boundary_bottom: int,
        material_name: str,
        head_line_top: int,
        head_line_bottom: int,
        overwrite: bool = False,  # TODO overwrite if layer already exists
    ) -> int:
        """Create layer based on boundary ids.

        .. todo::
            Determine how a 1D geometry would fit in here.
        """

        # TODO:: add check that the soil name exist
        # soilname = self.soils.add_soil(material)
        layer = Layer(
            material=material_name,
            piezo_top=head_line_top,
            piezo_bottom=head_line_bottom,
            boundary_top=boundary_top,
            boundary_bottom=boundary_bottom,
        )
        layer = self.layers.add_layer(layer)
        return layer.id

    # TODO check whether this needs to be done
    # def set_limits(self, x_min: float, x_max: float):
    #     """Set limits of geometry.

    #     .. todo::
    #         Determine how to handle points/layers outside of limits.
    #     """

    # 1.2.2 Loads
    @property
    def other_loads(self):
        """Enables easy access to the other loads in the internal dict-like datastructure."""
        return self.datastructure.other_loads

    def add_other_load(
        self,
        name: constr(min_length=1, max_length=25),
        time: timedelta,
        point: Point,
        other_load: Union[
            TrapeziformLoad, RectangularLoad, CircularLoad, TankLoad, UniformLoad
        ],
    ) -> None:
        internal_other_load = other_load._to_internal(time, point)
        if isinstance(self.other_loads, str):
            logger.warning("Replacing unparsed OtherLoads!")
            self.datastructure.other_loads = OtherLoads()
        self.other_loads.add_load(name, internal_other_load)

    @property
    def non_uniform_loads(self):
        """Enables easy access to the non-uniform loads in the internal dict-like datastructure."""
        return self.datastructure.non__uniform_loads

    def add_non_uniform_load(
        self,
        name: constr(min_length=1, max_length=25),
        points: List[Point],
        time_start: timedelta,
        gamma_dry: float,
        gamma_wet: float,
        time_end: Optional[timedelta] = None,
    ):
        """Create non uniform load.

        Sequence of loading is based on time_start.
        """
        # If end time is determined in D-Settlement then temporary value is True
        if time_end is None:
            time_end = timedelta(days=0)
            temporary = False
        else:
            temporary = True

        # List of points should be converted for the internal part of the code
        points_for_load = [PointForLoad.from_point(point) for point in points]

        non_uniform_load = NonUniformLoad(
            time=time_start.days,
            endtime=time_end.days,
            gammadry=gamma_dry,
            gammawet=gamma_wet,
            temporary=temporary,
            points=points_for_load,
        )

        if isinstance(self.non_uniform_loads, str):
            logger.warning("Replacing unparsed NonUniformLoads!")
            self.datastructure.non__uniform_loads = NonUniformLoads()
        self.non_uniform_loads.add_load(name, non_uniform_load)

    def add_water_load(self, name: str, time: timedelta, phreatic_line_id: int):
        """Create water load for a time in days, based on a phreatic line.

        Edit the head lines for each layer with `create layer`.
        """
        if isinstance(self.datastructure.water_loads, str):
            logger.warning("Replacing unparsed [WATER LOADS]!")
            self.datastructure.water_loads = WaterLoads()
        headlines = self.datastructure.get_headlines_for_layers()
        self.datastructure.water_loads.add_waterload(
            name, time.days, phreatic_line_id, headlines
        )

    def set_calculation_times(self, time_steps: List[timedelta]):
        """(Re)set calculation time(s).

        Sets a list of calculation times, sorted from low to high with a minimum of 0.

        Args:
            time_steps: List of time steps, type: float >= 0

        Returns:

        """
        time_steps.sort()
        residual_times = ResidualTimes(
            time_steps=[timestep.days for timestep in time_steps]
        )
        self.datastructure.residual_times = residual_times

    def set_verticals(self, locations: List[Point]) -> None:
        """
        Set calculation verticals in geometry.
        X and Y coordinates should be defined for each vertical.

        .. todo::
            Add check that checks that the verticals are not outside of the geometry boundaries. [GEOLIB-12]
        """
        points = [DSeriePoint.from_point(point) for point in locations]
        verticals = Verticals(locations=points)
        self.datastructure.verticals = verticals

    def set_vertical_drain(self, verticaldrain: VerticalDrain):
        if self.datastructure.model.is_vertical_drains:
            self.datastructure.vertical_drain = verticaldrain._to_internal()
        else:
            raise ValueError(
                "If you wish to add a vertical drain then value is_vertical_drains for the model should be True"
            )

    @property
    def output(self) -> Results:
        """Access internal dict-like datastructure of the output.

        Requires a successful execute.
        """
        return self.datastructure.results



import csv
import logging
import re
from collections import namedtuple
from pathlib import Path
from typing import Any

from pydantic import validator

_CAMEL_TO_SNAKE_PATTERN = re.compile(r"(?<!^)(?=[A-Z])")

logger = logging.getLogger(__name__)


def camel_to_snake(name: str) -> str:
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return _CAMEL_TO_SNAKE_PATTERN.sub("_", name).lower()


def snake_to_camel(name: str) -> str:
    return "".join(word.title() for word in name.split("_"))


def csv_as_namedtuples(fn: Path, delimiter=";"):
    with open(fn, newline="") as f:
        reader = csv.reader(f, delimiter=delimiter)
        header = namedtuple("Header", next(reader))
        for data in map(header._make, reader):
            yield data


def make_newline_validator(*field_name: str, req_newlines: int = 2):
    """Get a dynamic validator for ensuring a set number of lines."""

    def field_must_contain_newlines(v: str):
        newlines = v.count("\n")
        if newlines < req_newlines:
            logger.warning(
                f"Added {req_newlines - newlines} lines to run_identification."
            )
            v += (req_newlines - newlines) * "\n"
        elif newlines > req_newlines:
            logger.warning(
                f"More than {req_newlines+1} lines in run_identification will be ignored in the GUI."
            )
        return v

    return validator(*field_name, allow_reuse=True)(field_must_contain_newlines)



from typing import Optional

from geolib.models import BaseDataClass

from .internal import WaterLevel as InternalWaterLevel
from .settings import DistributionType


class WaterLevel(BaseDataClass):
    name: str
    level: float
    distribution_type: DistributionType = DistributionType.NONE
    standard_deviation: float = 0.0

    def to_internal(self) -> InternalWaterLevel:
        return InternalWaterLevel(**self.dict(exclude_none=True))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['geolib',
 'geolib.geometry',
 'geolib.models',
 'geolib.models.dfoundations',
 'geolib.models.dgeoflow',
 'geolib.models.dsettlement',
 'geolib.models.dsheetpiling',
 'geolib.models.dstability',
 'geolib.service',
 'geolib.soils']

package_data = \
{'': ['*'],
 'geolib.models': ['dfoundations/soil_csv/*',
                   'dfoundations/templates/*',
                   'dsettlement/templates/*',
                   'dsheetpiling/templates/*']}

install_requires = \
['jinja2>=3.1.2,<4.0.0',
 'pydantic[dotenv]>=1.6.1,<2.0.0',
 'requests>=2.28.1,<3.0.0',
 'zipp>=3.11.0,<4.0.0']

extras_require = \
{'server': ['fastapi>=0.88.0,<0.89.0',
            'uvicorn>=0.20.0,<0.21.0',
            'httpx>=0.23.1,<0.24.0']}

entry_points = \
{'console_scripts': ['geolib_server = geolib.service.main:app']}

setup_kwargs = {
    'name': 'd-geolib',
    'version': '0.3.0',
    'description': 'Python wrappers around the input and output files of the Deltares D-Serie and D-GEO Suite models',
    'long_description': "GEOLib\n=============================\n\nGEOLib is a Python package to generate, execute and parse several D-Serie and D-GEO Suite numerical models.\n\nInstallation\n------------\n\nInstall GEOLib with:\n\n.. code-block:: bash\n\n    $ pip install d-geolib\n\nConfigure your environment using the instructions on our `Setup <https://deltares.github.io/GEOLib/latest/user/setup.html>`_ page.\nYou may get the console executables from the Deltares download portal, or in the case of the D-GEO Suite, you may copy the contents of the installation 'bin' directory to your console folder.\n\nRunning the source code\n-----------------------\n\nIf you want to make changes to GEOLib you can run the source code from GitHub directly on your local machine, \nplease follow the instructions below on how to set up your development environment using pip or poetry.\n\nYou do not need to follow these instructions if you want to use the GEOLib package in your project.\n\nRequirements\n------------\n\nTo install the required dependencies to run GEOLib code, run:\n\n.. code-block:: bash\n\n    $ pip install -r requirements.txt\n\nOr, when having poetry installed (you should):\n\n.. code-block:: bash\n\n    $ poetry install\n\n\nTesting & Development\n---------------------\n\nMake sure to have the server dependencies installed: \n\n.. code-block:: bash\n\n    $ poetry install -E server\n\nIn order to run the testcode, from the root of the repository, run:\n\n.. code-block:: bash\n\n    $ pytest\n\nor, in case of using Poetry\n\n.. code-block:: bash\n\n    $ poetry run pytest\n\nRunning flake8, mypy is also recommended. For mypy use:\n\n.. code-block:: bash\n\n    $ mypy --config-file pyproject.toml geolib\n\n\nDocumentation\n-------------\n\nIn order to run the documentation, from the root of the repository, run:\n\n.. code-block:: bash\n\n    $ cd docs\n    $ sphinx-build . build -b html -c .\n\n\nThe documentation is now in the `build` subfolder, where you can open \nthe `index.html` in your browser.\n\nBuild wheel\n-----------\n\nTo build a distributable wheel package, run:\n\n.. code-block:: bash\n\n    $ poetry build\n\nThe distributable packages are now built in the `dist` subfolder.\n\nUpdate requirements.txt\n-----------------------\n\nThe requirements.txt file is generated by poetry based on the pyproject.toml and poetry.lock files. In order to update/regenerate this file, run:\n\n.. code-block:: bash\n\n    $ poetry install\n    $ poetry export -E server -f requirements.txt --output requirements.txt --without-hashes\n    $ poetry export -E server -f requirements.txt --output requirements-dev.txt --with dev --without-hashes\n",
    'author': 'Maarten Pronk',
    'author_email': 'maarten.pronk@deltares.nl',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://deltares.github.io/GEOLib/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from abc import ABCMeta, abstractmethod
from typing import Optional, Union

from pydantic.types import confloat, conint

from geolib.models import BaseDataClass

from .settings import (
    CalculationType,
    DesignType,
    PartialFactorCalculationType,
    PartialFactorSetCUR,
    PartialFactorSetEC,
    PartialFactorSetEC7NADB,
    PartialFactorSetEC7NADNL,
    PartialFactorSetVerifyEC,
    VerifyType,
)


class CalculationOptionsPerStage(BaseDataClass):
    """Calculation options are needed when a verification calculation is performed with method B.
    The user can specify a different partial factor set for each of the stages. This input is optional
    when another type of calculation is performed.
    Arguments:
     anchor_factor: multiplication factor for the anchor stiffness.
     partial_factor_set: partial factor sets corresponding to different code annexes.
    """

    anchor_factor: float = 1
    partial_factor_set: Optional[
        Union[PartialFactorSetEC7NADB, PartialFactorSetCUR, PartialFactorSetEC7NADNL]
    ] = None


class CalculationOptions(BaseDataClass, metaclass=ABCMeta):
    """Base class for all calculation options.

    Arguments:
     calc_first_stage_initial: This option determines equal neutral stresses at both sides, for initially non-horizontal surfaces or initial surcharges.
     calc_minor_nodes_on: Select either the faster, classic, coarse element determination (False) of active and passive pressures, or the more accurate fine element determination(True).
     calc_reduce_deltas:  Set on True for reduction of the wall friction angles according to CUR 166 for the calculation of the passive earth pressure coefficients of Culmann method.
     input_calculation_type: select the type of calculation that is going to be performed.

    """

    calc_first_stage_initial: bool = False
    calc_minor_nodes_on: bool = False
    calc_reduce_deltas: bool = False  # editable if C,phi, delta is selected
    input_calculation_type: CalculationType = CalculationType.STANDARD
    is_vibration_calculation: bool = False

    @property
    def calculation_properties(
        self,
    ) -> Union[
        "StandardCalculationOptions",
        "DesignSheetpilingLengthCalculationOptions",
        "VerifyCalculationOptions",
        "KranzAnchorStrengthCalculationOptions",
        "OverallStabilityCalculationOptions",
        "ReliabilityAnalysisCalculationOptions",
    ]:
        _calculation_properties_map = {
            CalculationType.STANDARD: StandardCalculationOptions,
            CalculationType.DESIGN_SHEETPILING_LENGTH: DesignSheetpilingLengthCalculationOptions,
            CalculationType.VERIFY_SHEETPILING: VerifyCalculationOptions,
            CalculationType.CHARACTERISTIC_KRANZ_ANCHOR_STRENGTH: KranzAnchorStrengthCalculationOptions,
            CalculationType.OVERALL_STABILITY: OverallStabilityCalculationOptions,
            CalculationType.RELIABILITY_ANALYSIS: ReliabilityAnalysisCalculationOptions,
        }
        return _calculation_properties_map[self.input_calculation_type]()


class ReliabilityAnalysisCalculationOptions(CalculationOptions):
    """Reliability analysis calculation selected in Start Calculation
    window."""

    input_calculation_type: CalculationType = CalculationType.RELIABILITY_ANALYSIS


class OverallStabilityCalculationOptions(CalculationOptions):
    """Overall stability calculation selected in Start Calculation window.

    Arguments:
     cur_stability_stage: Id of the stage to be checked. This id refers to D-SheetPiling so the first stage in D-SheetPiling has an input of 0.
     overall_stability_type: which type of calculation is going to be performed
     stability_eurocode_partial_factor_set: partial factor set selected for the EC7 General calculation
     stability_ec7_nl_partial_factor_set: partial factor set selected for the EC7 NL calculation
     stability_ec7_b_partial_factor_set: partial factor set selected for the EC7 B calculation
     stability_cur_partial_factor_set: partial factor set selected for the CUR calculation
    """

    input_calculation_type: CalculationType = CalculationType.OVERALL_STABILITY
    cur_stability_stage: conint(ge=0) = 0
    overall_stability_type: DesignType = DesignType.REPRESENTATIVE
    stability_eurocode_partial_factor_set: PartialFactorSetEC = PartialFactorSetEC.DA1SET1
    stability_ec7_nl_partial_factor_set: PartialFactorSetEC7NADNL = (
        PartialFactorSetEC7NADNL.RC0
    )
    stability_ec7_b_partial_factor_set: PartialFactorSetEC7NADB = (
        PartialFactorSetEC7NADB.SET1
    )
    stability_cur_partial_factor_set: PartialFactorSetCUR = PartialFactorSetCUR.CLASSI


class KranzAnchorStrengthCalculationOptions(CalculationOptions):
    """Kranz anchor strength calculation selected in Start Calculation
    window.

    Arguments:
     cur_anchor_force_stage: Id of the stage to be checked. This id refers to D-SheetPiling so the first stage in D-SheetPiling has an input of 0.
    """

    input_calculation_type: CalculationType = (
        CalculationType.CHARACTERISTIC_KRANZ_ANCHOR_STRENGTH
    )
    cur_anchor_force_stage: conint(ge=0) = 0


class StandardCalculationOptions(CalculationOptions):
    """Standard calculation selected in Start Calculation window.

    Arguments:
     calc_auto_lambdas_on: When True Automatic leaves the values of the lateral earth pressure ratios that are calculated by the Culmann (c, phi, delta) method as they are.
    """

    input_calculation_type: CalculationType = CalculationType.STANDARD
    calc_auto_lambdas_on: bool = True


class VerifyCalculationOptions(CalculationOptions):
    """Verify sheet pile calculation selected in Start Calculation window.

    Arguments:
     verify_type: Select partial factor set
     eurocode_partial_factor_set: Select partial factor set
     eurocode_overall_stability: Set to True to perform an overall stability calculation using modified values for soil properties (cohesion, friction angle and unit weight) depending on the Design approach chosen for all stages.
     ec7_nl_method: Select method of calculation according to CUR 166 design procedure
     ec7_nl_overall_partial_factor_set: Select partial factor set
     ec7_nl_overall_anchor_factor: multiplication factor for the anchor stiffness
     ec7_nad_nl_overall_stability: Set to True to perform an overall stability calculation using modified values for soil properties (cohesion, friction angle and unit weight) depending on the Design approach chosen for all stages.
     ec7_b_overall_stability: Set to True to perform an overall stability calculation using modified values for soil properties (cohesion, friction angle and unit weight) depending on the Design approach chosen for all stages.
     nb_method: Select method of calculation according to CUR 166 design procedure
     cur_method: Select method of calculation according to CUR 166 design procedure
     cur_overall_partial_factor_set: Select partial factor set
     cur_overall_anchor_factor: multiplication factor for the anchor stiffness
     cur_overall_stability: Set to True to perform an overall stability calculation using modified values for soil properties (cohesion, friction angle and unit weight) depending on the Design approach chosen for all stages.
    """

    input_calculation_type: CalculationType = CalculationType.VERIFY_SHEETPILING
    verify_type: VerifyType = VerifyType.CUR
    eurocode_partial_factor_set: PartialFactorSetVerifyEC = PartialFactorSetVerifyEC.DA1
    eurocode_overall_stability: bool = False
    ec7_nl_method: PartialFactorCalculationType = PartialFactorCalculationType.METHODA
    ec7_nl_overall_partial_factor_set: PartialFactorSetEC7NADNL = (
        PartialFactorSetEC7NADNL.RC0
    )
    ec7_nl_overall_anchor_factor: confloat(ge=0.001, le=1000) = 1
    ec7_nad_nl_overall_stability: bool = False
    ec7_b_overall_stability: bool = False
    nb_method: PartialFactorCalculationType = PartialFactorCalculationType.METHODA
    cur_method: PartialFactorCalculationType = PartialFactorCalculationType.METHODA
    cur_overall_partial_factor_set: PartialFactorSetCUR = PartialFactorSetCUR.CLASSI
    cur_overall_anchor_factor: confloat(ge=0.001, le=1000) = 1
    cur_overall_stability: bool = False

    @property
    def allowable_anchor_force_calculation_type(self) -> bool:
        return (self.verify_type == VerifyType.CUR) or (
            self.verify_type == VerifyType.EC7NL
        )


class DesignSheetpilingLengthCalculationOptions(CalculationOptions):
    """Design sheet pile calculation selected in Start Calculation window.

    Note that design_pile_length can show up as a different value in the GUI
    compared to what's been set in the .shi file.

    Args:
     design_stage: Id of the stage to be checked. This id refers to D-SheetPiling so the first stage in D-SheetPiling has an input of 0.
     design_pile_length_from: The starting point of the pile over which the analysis should be performed.
     design_pile_length_to: The end point of the pile over which the analysis should be performed
     design_pile_length_decrement: the Decrement in length for each analysis step.
     design_type: types of design.
     design_eurocode_partial_factor_set: Select partial factor set
     design_partial_factor_set_ec7_nad_nl: Select partial factor set
     design_ec7_nl_method: Select method of calculation according to CUR 166 design procedure
     design_partial_factor_set_ec7_nad_b: Select partial factor set
     design_ec7_b_method: Select method of calculation according to CUR 166 design procedure
     design_partial_factor_set: Select partial factor set
     design_cur_method: Select method of calculation according to CUR 166 design procedure
    """

    input_calculation_type: CalculationType = CalculationType.DESIGN_SHEETPILING_LENGTH
    design_stage: conint(ge=0) = 0
    design_pile_length_from: confloat(ge=1, le=100) = 1
    design_pile_length_to: confloat(ge=1, le=100) = 1
    design_pile_length_decrement: confloat(ge=0.01, le=10) = 0.01
    design_type: DesignType = DesignType.REPRESENTATIVE
    design_eurocode_partial_factor_set: PartialFactorSetEC = PartialFactorSetEC.DA1SET1
    design_partial_factor_set_ec7_nad_nl: PartialFactorSetEC7NADNL = (
        PartialFactorSetEC7NADNL.RC0
    )
    design_ec7_nl_method: PartialFactorCalculationType = (
        PartialFactorCalculationType.METHODA
    )
    design_partial_factor_set_ec7_nad_b: PartialFactorSetEC7NADB = (
        PartialFactorSetEC7NADB.SET1
    )
    design_ec7_b_method: PartialFactorCalculationType = (
        PartialFactorCalculationType.METHODA
    )
    design_partial_factor_set: PartialFactorSetCUR = PartialFactorSetCUR.CLASSI
    design_cur_method: PartialFactorCalculationType = PartialFactorCalculationType.METHODA



from typing import List, Tuple, Type

from pydantic import FilePath

from geolib.models.dseries_parser import DSerieParser, DSeriesStructure
from geolib.models.parsers import BaseParserProvider

from .internal import DSheetPilingDumpStructure, DSheetPilingStructure


class DSheetPilingInputParser(DSerieParser):
    """DSheetPiling parser of input files."""

    @property
    def suffix_list(self) -> List[str]:
        return [".shi"]

    @property
    def dserie_structure(self) -> Type[DSheetPilingStructure]:
        return DSheetPilingStructure


class DSheetPilingOutputParser(DSerieParser):
    """DSheetPiling parser of input files."""

    @property
    def suffix_list(self) -> List[str]:
        return [".shd"]

    @property
    def dserie_structure(self) -> Type[DSheetPilingDumpStructure]:
        return DSheetPilingDumpStructure


class DSheetPilingParserProvider(BaseParserProvider):

    __input_parsers = None
    __output_parsers = None

    @property
    def input_parsers(self) -> Tuple[DSheetPilingInputParser]:
        if not self.__input_parsers:
            self.__input_parsers = (DSheetPilingInputParser(),)
        return self.__input_parsers

    @property
    def output_parsers(self) -> Tuple[DSheetPilingOutputParser]:
        if not self.__output_parsers:
            self.__output_parsers = (DSheetPilingOutputParser(),)
        return self.__output_parsers

    @property
    def parser_name(self) -> str:
        return "DSheetPiling"



from abc import ABCMeta, abstractmethod
from pathlib import Path
from subprocess import CompletedProcess, run
from typing import Any, List, Optional, Type, Union
from typing import BinaryIO

from pydantic import FilePath, PositiveFloat
from pydantic.types import confloat, conint

from geolib.geometry import Point
from geolib.models import BaseDataClass, BaseModel, BaseModelStructure
from geolib.models.dsheetpiling.constructions import DiaphragmWall, Pile, Sheet
from geolib.models.meta import CONSOLE_RUN_BATCH_FLAG
from geolib.soils import Soil

from .calculation_options import CalculationOptions, CalculationOptionsPerStage
from .dsheetpiling_parserprovider import DSheetPilingParserProvider
from .internal import CalculationOptions as CalculationOptionsInternal
from .internal import (
    DSheetPilingDumpStructure,
    DSheetPilingOutputStructure,
    DSheetPilingStructure,
    Water,
)
from .loads import (
    Earthquake,
    HorizontalLineLoad,
    Moment,
    NormalForce,
    SoilDisplacement,
    SurchargeLoad,
    UniformLoad,
)
from .profiles import SoilProfile
from .serializer import DSheetPilingInputSerializer
from .settings import (
    CalculationType,
    CurveSettings,
    DesignType,
    LateralEarthPressureMethod,
    LateralEarthPressureMethodStage,
    ModelType,
    PartialFactorCalculationType,
    PartialFactorSetCUR,
    PartialFactorSetEC,
    PartialFactorSetEC7NADB,
    PartialFactorSetEC7NADNL,
    PartialFactorSetVerifyEC,
    PassiveSide,
    Side,
    SinglePileLoadOptions,
    VerifyType,
)
from .supports import Anchor, RigidSupport, SpringSupport, Strut
from .surface import Surface
from .water_level import WaterLevel


class BaseModelType(BaseDataClass, metaclass=ABCMeta):
    """Base Class for Model types (Pile, Sheet, Wall)."""

    @property
    @abstractmethod
    def model(self):
        raise NotImplementedError

    def to_internal(self):
        return dict(self)


class SheetModelType(BaseModelType):
    method: Optional[LateralEarthPressureMethod] = None
    check_vertical_balance: Optional[bool] = None
    verification: Optional[bool] = None
    trildens_calculation: Optional[bool] = None
    reliability_analysis: Optional[bool] = None
    elastic_calculation: Optional[bool] = None

    @property
    def model(self) -> ModelType:
        return ModelType.SHEET_PILING


class WoodenSheetPileModelType(BaseModelType):
    method: Optional[LateralEarthPressureMethod] = None
    check_vertical_balance: Optional[bool] = None
    verification: Optional[bool] = None
    elastic_calculation: Optional[bool] = None

    @property
    def model(self) -> ModelType:
        return ModelType.SHEET_PILING


class SinglePileModelType(BaseModelType):
    pile_load_option: Optional[SinglePileLoadOptions] = None
    elastic_calculation: Optional[bool] = None

    @property
    def model(self) -> ModelType:
        return ModelType.SINGLE_PILE

    def to_internal(self):
        if self.pile_load_option:
            if self.pile_load_option.value == "forces":
                return {"pile_load_option": 0, "pile_load_by_user": 0}
            elif self.pile_load_option.value == "user_defined_displacements":
                return {"pile_load_option": 1, "pile_load_by_user": 0}
            elif self.pile_load_option.value == "calculated_displacements":
                return {"pile_load_option": 1, "pile_load_by_user": 1}
        else:
            return {}


class DiaphragmModelType(BaseModelType):
    method: Optional[LateralEarthPressureMethod] = None
    check_vertical_balance: Optional[bool] = None
    verification: Optional[bool] = None

    @property
    def model(self) -> ModelType:
        return ModelType.DIAPHRAGM_WALL


class DSheetPilingModel(BaseModel):
    r"""DSheetPiling is a tool used to design sheetpile and diaphragm walls and
    horizontally loaded piles.

    This model can read, modify and create \*.shi files, read \*.shd and \*.err files.
    """

    current_stage: Optional[int] = None  # Forces user to always set a stage.
    datastructure: Union[
        DSheetPilingStructure, DSheetPilingDumpStructure
    ] = DSheetPilingStructure()

    @property
    def parser_provider_type(self) -> Type[DSheetPilingParserProvider]:
        return DSheetPilingParserProvider

    @property
    def console_path(self) -> Path:
        return Path("DSheetPilingConsole/DSheetPilingConsole.exe")

    @property
    def console_flags(self) -> List[str]:
        return [CONSOLE_RUN_BATCH_FLAG]

    @property
    def output(self) -> DSheetPilingOutputStructure:
        return self.datastructure.dumpfile.output_data

    @property
    def model_type(self) -> Union[str, ModelType]:
        return self.datastructure.input_data.model.model

    def serialize(self, filename: Union[FilePath, BinaryIO]):
        ds = self.datastructure.input_data.dict()
        ds.update(
            {
                "version": self.datastructure.version.dict(),
                "version_externals": self.datastructure.version_externals.dict(),
            }
        )
        serializer = DSheetPilingInputSerializer(ds=ds)
        serializer.write(filename)
        if isinstance(filename, Path):
            self.filename = filename

    def _is_calculation_per_stage_required(self) -> bool:
        """Function that checks if [CALCULATION PER STAGE] can be modified. This is true for a verify sheet-piling calculation and method B."""
        _map_method_b_available = {
            VerifyType.CUR: self.datastructure.input_data.calculation_options.curmethod,
            VerifyType.EC7NL: self.datastructure.input_data.calculation_options.ec7nlmethod,
            VerifyType.EC7BE: self.datastructure.input_data.calculation_options.nbmethod,
        }
        if (
            self.datastructure.input_data.calculation_options.inputcalculationtype
            == CalculationType.VERIFY_SHEETPILING
        ):
            if (
                _map_method_b_available[
                    self.datastructure.input_data.calculation_options.verifytype
                ]
                == PartialFactorCalculationType.METHODA
            ):
                raise ValueError(
                    "MethodA selected for VERIFY_SHEETPILING calculation. Calculation_per_stage options should not be defined or the type of calculation should be switch to MethodB"
                )
        else:
            raise ValueError(
                "Calculation_per_stage should only be defined if VERIFY_SHEETPILING calculation was selected."
            )
        return True

    def set_unit_weight_water(self, unit_weight: float) -> None:
        """Sets the unit weight for water in the [WATER] block"""
        self.datastructure.input_data.water = Water(unit_weight_of_water=unit_weight)

    def add_stage(
        self,
        name: str,
        passive_side: PassiveSide,
        method_left: LateralEarthPressureMethodStage,
        method_right: LateralEarthPressureMethodStage,
        pile_top_displacement: float = 0.0,
    ) -> int:
        """Add a new stage to the model.

        When using a Single Pile model, the lateral earth pressure method left and right need to be the same.
        Their inputs however do not effect the outcome of the calculation.

        Args:
            name: Name of the stage.
            passive_side: Option to set the passive side for the stage.
            method_left: LateralEarthPressureMethodStage applied to left side, must be compatible with the Model LateralEarthPressureMethod
            method_right: LateralEarthPressureMethodStage applied to right side, must be compatible with the Model LateralEarthPressureMethod
            pile_top_displacement: Pile top displacement [m]. When not provided, Use top displacement will be set to false.

        Raises:
            ValidationError: when input arguments are not within constraints

        Returns:
            int: Stage id which can be used to modify the stage.
        """
        new_stage_id = self.current_stage + 1 if self.current_stage is not None else 0
        self.datastructure.input_data.add_stage(
            name, passive_side, method_left, method_right, pile_top_displacement
        )
        self.current_stage = new_stage_id
        return new_stage_id

    def add_calculation_options_per_stage(
        self, calculation_options_per_stage: CalculationOptionsPerStage, stage_id: int
    ) -> None:
        """Set calculation options per stage.

        Calculation options per stage are set in [CALCULATION OPTIONS PER STAGE].

        Args:
            calculation_options_per_stage: Calculation options for stage
            stage_id: Settings are set to this stage. This refers to the Pythonic input and has a starting point of 0.

        Raises:
            ValueError: When non-existing stage_id is passed or when no
                        CalculationOptionsPerStage are required.
        """
        if self._is_calculation_per_stage_required():
            self._check_if_stage_id_exists(stage_id)
            self.datastructure.input_data.add_calculation_options_per_stage(
                input_calc_options=calculation_options_per_stage,
                stage_id=stage_id,
            )
        else:
            raise ValueError(
                f"calculation_options_per_stage is not required for stage {stage_id}."
            )

    def _get_default_stage_none_provided(self, stage_id: Optional[int]) -> int:
        stage_id = stage_id if stage_id is not None else self.current_stage
        if stage_id is None:
            raise ValueError("No stages available yet: add a stage first.")
        self._check_if_stage_id_exists(stage_id)
        return stage_id

    def _check_if_stage_id_exists(self, stage_id: int) -> None:
        if not self.datastructure.input_data.has_stage(stage_id):
            raise ValueError(
                f"Stage {stage_id} is not added to the internal datastructure"
            )

    def set_model(self, model: BaseModelType) -> None:
        """(Re)Set BaseModelType (Pile, Sheet, Wall) for model."""
        if not issubclass(type(model), BaseModelType):
            raise ValueError(
                f"model should be of subtype BaseModelType, received {model}"
            )

        self.datastructure.input_data.set_model(**model.to_internal(), model=model.model)

    def set_calculation_options(self, calculation_options: CalculationOptions) -> None:
        """Set calculation options.

        Overall calculation options are set in [CALCULATION OPTIONS].

        Args:
            calculation_options: Calculation options
        """
        if not issubclass(type(calculation_options), CalculationOptions):
            raise ValueError(
                f"model should be of subtype CalculationOptions, received {calculation_options}"
            )
        self.datastructure.input_data.set_calculation_options(**dict(calculation_options))

    def set_curve_settings(
        self,
        curve_settings: CurveSettings,
    ):
        """Set curve settings for soil profiles.

        Curve settings are set in  [SOIL PROFILES].

        Args:
            curve_settings: Curvesettings
        """
        self.datastructure.input_data.set_curve_settings(curve_settings)

    def add_head_line(
        self,
        water_level: WaterLevel,
        side: Side,
        stage_id: int,
    ) -> None:
        """Set water level for a stage.

        If a water level already exists, it is refered to that water level.

        Related to the [WATERLEVELS] block in the .shi file.

        Args:
            water_level: WaterLevel.
            side: Side which determines on which side the surface is added.
            stage_id: WaterLevel is added to this stage.

        Raises:
            ValueError: When non-existing stage_id is passed.
        """
        self._check_if_stage_id_exists(stage_id)
        self.datastructure.input_data.add_water_level(
            stage_id, water_level.to_internal(), side=side
        )

    def add_surface(
        self,
        surface: Surface,
        side: Side,
        stage_id: int,
    ) -> None:
        """Set surface for a stage.

        Surface is added to [SURFACES] if not yet added; reference is done by name.
        A reference in [CONSTRUCTION STAGES] is updated.

        Args:
            surface: Surface.
            side: Side which determines on which side the surface is added.
            stage_id: Surface is added to this stage.

        Raises:
            ValueError: When non-existing stage_id is passed.
        """
        self._check_if_stage_id_exists(stage_id)
        self.datastructure.input_data.add_surface(
            stage_id, surface.to_internal(), side=side
        )

    def add_profile(
        self,
        profile: SoilProfile,
        side: Side,
        stage_id: int,
    ) -> None:
        """Add a Profile on the left or right side of a stage.

        Profile is added to [SOIL PROFILES] if not yet added; reference is done by name.
        A reference in [CONSTRUCTION STAGES] is updated.

        Args:
            profile: Profile.
            side: Side which determines on which side the profile is added.
            stage_id: Surface is added to this stage.

        Raises:
            ValueError: When non-existing stage_id is passed.
        """
        self._check_if_stage_id_exists(stage_id)
        self.datastructure.input_data.add_profile(
            stage_id, profile.to_internal(), side=side
        )

    def set_construction(
        self, top_level: float, elements: List[Union[Sheet, DiaphragmWall, Pile]]
    ) -> None:
        """Sets construction for the DSheetPilingModel.

        Elements are added to [SHEET PILING].
        Removes current sheetpiling when called.

        Args:
            top_level: Top level of the sheet piling.
            elements: List of sheet piling elements (can be Sheet, DiaphragmWall, or Pile elements) which are sorted on sheetpilingelementlevel.
        """
        self.datastructure.input_data.set_construction(
            top_level=top_level, elements=[element.to_internal() for element in elements]
        )

    def add_load(
        self,
        load: Union[
            Moment,
            HorizontalLineLoad,
            NormalForce,
            SoilDisplacement,
            Earthquake,
            UniformLoad,
        ],
        stage_id: int,
    ):
        """Adds other loads of type Moment, HorizontalLineLoad, NormalForce, SoilDisplacement or Earthquake

        Args:
            load: Add a load with the types of Moment, HorizontalLineLoad, NormalForce, SoilDisplacement or Earthquake.
            Note that SoilDisplacement and Earthquake are only valid for a Pile construction.
            stage_id: Load is added to this stage.

        Raises:
            ValueError: When non-existing stage_id is passed.
            ValueError: When a verification calculation is selected but duration_type and load_type are not defined for the load.
        """
        self._check_if_stage_id_exists(stage_id)
        self.datastructure.input_data.add_load(load=load.to_internal(), stage_id=stage_id)

    def add_surcharge_load(self, load: SurchargeLoad, side: Side, stage_id: int) -> None:
        """Add surcharge load to a stage.

        Args:
            load: SurchargeLoad.
            side: Side which determines on which side the load is added.
            stage_id: SurchareLoad is added to this stage.

        Raises:
            ValueError: When non-existing stage_id is passed.
        """
        self._check_if_stage_id_exists(stage_id)
        self.datastructure.input_data.add_surcharge_load(
            stage_id, load.to_internal(), side=side
        )

    def add_anchor_or_strut(
        self,
        support: Union[Anchor, Strut],
        stage_id: int,
        pre_stress: PositiveFloat = 0,
    ) -> None:
        """Add anchor or strut to a stage.

        Anchor and Strut supports are only available for the sheetpiling model.

        Args:
            support: Anchor and Strut.
            pre_stress: Prestress for the support which is added to a construction stage.
                        This is a tension stress for the anchor and a compression stress for the strut.
            stage_id: Support is added to this stage.
        """
        self._check_if_stage_id_exists(stage_id)

        if isinstance(self.model_type, str):
            raise ValueError("Cannot read model type when it's not yet set; set model.")
        if self.model_type == ModelType.SINGLE_PILE:
            raise ValueError(
                f"Only spring and rigid supports can be used for the single pile model; received {support}"
            )

        if isinstance(support, Anchor):
            self.datastructure.input_data.add_anchor(
                stage_id, support.to_internal(), pre_tension=pre_stress
            )
        elif isinstance(support, Strut):
            self.datastructure.input_data.add_strut(
                stage_id, support.to_internal(), pre_compression=pre_stress
            )
        else:
            raise ValueError(f"support should be Anchor or Strut, received {support}")

    def add_support(
        self,
        support: Union[SpringSupport, RigidSupport],
        stage_id: int,
    ) -> None:
        """Add spring or rigid support to a stage.

        Args:
            support: SpringSupport or RigidSupport.
            stage_id: ID of the stage. Of no ID is provided, the current stage ID will be taken.
        """
        self._check_if_stage_id_exists(stage_id)
        if isinstance(support, SpringSupport):
            self.datastructure.input_data.add_spring_support(
                stage_id,
                support.to_internal(),
            )
        elif isinstance(support, RigidSupport):
            self.datastructure.input_data.add_rigid_support(
                stage_id,
                support.to_internal(),
            )
        else:
            raise ValueError(
                f"support should be SpringSupport or RigidSupport, received {support}"
            )

    def add_soil(self, soil: Soil) -> str:
        """Soil is converted in the internal structure and added in soil_collection."""
        soil_internal = soil._to_dsheetpiling()
        new_soil = self.datastructure.input_data.soil_collection.add_soil_if_unique(
            soil_internal
        )
        return new_soil.name

    @property
    def sheets(self):
        """Enables easy access to the sheets in the internal dict-like
        datastructure.

        Also enables edit/delete for individual Sheets.
        """

    @property
    def profiles(self):
        """Enables easy access to the profiles in the internal dict-like
        datastructure.

        Also enables edit/delete for individual Profiles.
        """



from typing import List, Tuple, Type

from pydantic import FilePath

from geolib.models.dseries_parser import DSerieParser, DSeriesStructure
from geolib.models.parsers import BaseParserProvider

from .internal import (
    DFoundationsDumpfileOutputStructure,
    DFoundationsDumpStructure,
    DFoundationsInputStructure,
    DFoundationsStructure,
)


class DFoundationsInputParser(DSerieParser):
    """DFoundations parser of input files."""

    @property
    def suffix_list(self) -> List[str]:
        return [".foi"]

    @property
    def dserie_structure(self) -> Type[DFoundationsStructure]:
        return DFoundationsStructure


class DFoundationsOutputParser(DSerieParser):
    """DFoundations parser of input files."""

    @property
    def suffix_list(self) -> List[str]:
        return [".fod"]

    @property
    def dserie_structure(self) -> Type[DFoundationsDumpStructure]:
        return DFoundationsDumpStructure


class DFoundationsParserProvider(BaseParserProvider):

    __input_parsers = None
    __output_parsers = None

    @property
    def input_parsers(self) -> Tuple[DFoundationsInputParser]:
        if not self.__input_parsers:
            self.__input_parsers = (DFoundationsInputParser(),)
        return self.__input_parsers

    @property
    def output_parsers(self) -> Tuple[DFoundationsOutputParser]:
        if not self.__output_parsers:
            self.__output_parsers = (DFoundationsOutputParser(),)
        return self.__output_parsers

    @property
    def parser_name(self) -> str:
        return "DFoundations"
